#  Copyright 2007-2012 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__author__ = "Rajiv Mayani"


from sqlalchemy.orm.exc import NoResultFound

from Pegasus.db.workflow import stampede_statistics
from Pegasus.plots_stats import utils as stats_utils
from Pegasus.service.dashboard import queries
from Pegasus.tools import utils


class NoWorkflowsFoundError(Exception):
    def __init__(self, **args):
        if "count" in args:
            self.count = args["count"]
        else:
            self.count = 0

        if "filtered" in args:
            self.filtered = args["filtered"]


class Dashboard:
    def __init__(self, master_db_url, root_wf_id=None, wf_id=None):
        self._master_db_url = master_db_url

        # If the ID is specified, it means that the query is specific to a workflow.
        # So we will now query the master database to get the connection URL for the workflow.
        if root_wf_id or wf_id:
            self.initialize(root_wf_id, wf_id)

    def initialize(self, root_wf_id, wf_id):
        try:
            workflow = queries.MasterDatabase(self._master_db_url)
            self._db_id, self._root_wf_uuid, self._wf_db_url = workflow.get_wf_id_url(
                root_wf_id
            )
            self._wf_id = wf_id
        finally:
            Dashboard.close(workflow)

    @staticmethod
    def close(conn):
        if conn:
            conn.close()

    def __get_wf_db_url(self):
        if not self._wf_db_url:
            raise ValueError("workflow database URL is not set")

        return self._wf_db_url

    def get_root_workflow_list(self, counts_only=False, **table_args):
        """
        Get basic information about all workflows running, on all databases. This is for the index page.
        Returns a list of workflows.
        """
        self._workflows = []

        # Now, let's try to access the database
        try:
            all_workflows = None
            all_workflows = queries.MasterDatabase(self._master_db_url)
            counts = all_workflows.get_workflow_counts()

            if counts_only:
                if counts[0] == 0:
                    raise NoWorkflowsFoundError(count=None, filtered=None)

                return counts

            count, filtered, workflows = all_workflows.get_all_workflows(**table_args)

            if workflows:
                self._workflows.extend(workflows)

            if len(self._workflows) == 0:
                # Throw no workflows found error.
                raise NoWorkflowsFoundError(count=count, filtered=filtered)

            return (count, filtered, self._workflows, counts)

        finally:
            Dashboard.close(all_workflows)

    def workflow_stats(self):
        try:
            workflow = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url(), False
            )
            workflow.initialize(root_wf_id=self._wf_id)
            individual_stats = self._workflow_stats(workflow)

            workflow2 = stampede_statistics.StampedeStatistics(self.__get_wf_db_url())
            workflow2.initialize(self._root_wf_uuid)

            all_stats = self._workflow_stats(workflow2)

            return {"individual": individual_stats, "all": all_stats}

        finally:
            Dashboard.close(workflow)
            Dashboard.close(workflow2)

    def _workflow_stats(self, workflow):
        # tasks
        tasks = {}
        workflow.set_job_filter("nonsub")
        tasks["total_tasks"] = int(workflow.get_total_tasks_status())
        tasks["total_succeeded_tasks"] = int(
            workflow.get_total_succeeded_tasks_status(False)
        )
        tasks["total_failed_tasks"] = int(workflow.get_total_failed_tasks_status())
        tasks["total_unsubmitted_tasks"] = tasks["total_tasks"] - (
            tasks["total_succeeded_tasks"] + tasks["total_failed_tasks"]
        )
        tasks["total_task_retries"] = int(workflow.get_total_tasks_retries())
        tasks["total_task_invocations"] = (
            tasks["total_succeeded_tasks"]
            + tasks["total_failed_tasks"]
            + tasks["total_task_retries"]
        )

        # job status
        jobs = {}
        workflow.set_job_filter("nonsub")
        jobs["total_jobs"] = int(workflow.get_total_jobs_status())
        jobs["total_succeeded_jobs"] = int(workflow.get_total_succeeded_jobs_status())
        jobs["total_failed_jobs"] = int(workflow.get_total_failed_jobs_status())
        jobs["total_unsubmitted_jobs"] = jobs["total_jobs"] - (
            jobs["total_succeeded_jobs"] + jobs["total_failed_jobs"]
        )
        jobs["total_job_retries"] = int(workflow.get_total_jobs_retries())
        jobs["total_job_invocations"] = (
            jobs["total_succeeded_jobs"]
            + jobs["total_failed_jobs"]
            + jobs["total_job_retries"]
        )

        # sub workflow
        wfs = {}
        workflow.set_job_filter("subwf")
        wfs["total_sub_wfs"] = int(workflow.get_total_jobs_status())
        wfs["total_succeeded_sub_wfs"] = int(workflow.get_total_succeeded_jobs_status())
        wfs["total_failed_sub_wfs"] = int(workflow.get_total_failed_jobs_status())
        wfs["total_unsubmitted_sub_wfs"] = wfs["total_sub_wfs"] - (
            wfs["total_succeeded_sub_wfs"] + wfs["total_failed_sub_wfs"]
        )
        wfs["total_sub_wfs_retries"] = int(workflow.get_total_jobs_retries())
        wfs["total_sub_wfs_invocations"] = (
            wfs["total_succeeded_sub_wfs"]
            + wfs["total_failed_sub_wfs"]
            + wfs["total_sub_wfs_retries"]
        )

        return [tasks, jobs, wfs]

    def job_breakdown_stats(self):
        try:
            workflow = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url(), True
            )
            workflow.initialize(root_wf_id=self._wf_id)
            content = []
            for t in workflow.get_transformation_statistics():
                content.append(
                    [
                        t.transformation,
                        t.type,
                        int(t.count),
                        "{:.2f}".format(t.min),
                        "{:.2f}".format(t.max),
                        "{:.2f}".format(t.avg),
                        "{:.2f}".format(t.sum),
                        "{:.2f}".format(t.min_maxrss / 1024) if t.min_maxrss else "-",
                        "{:.2f}".format(t.max_maxrss / 1024) if t.max_maxrss else "-",
                        "{:.2f}".format(t.avg_maxrss / 1024) if t.avg_maxrss else "-",
                        "{:.2f}%".format(t.min_avg_cpu * 100) if t.min_avg_cpu else "-",
                        "{:.2f}%".format(t.max_avg_cpu * 100) if t.max_avg_cpu else "-",
                        "{:.2f}%".format(t.avg_avg_cpu * 100) if t.avg_avg_cpu else "-",
                    ]
                )

            return content

        finally:
            Dashboard.close(workflow)

    def job_stats(self):
        try:
            workflow = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url(), False
            )
            workflow.initialize(root_wf_id=self._wf_id)
            workflow.set_job_filter("all")

            job_retry_count_dict = {}
            content = []

            for job in workflow.get_job_statistics():

                kickstart = "0" if job.kickstart is None else float(job.kickstart)
                multiplier_factor = (
                    "0" if job.multiplier_factor is None else int(job.multiplier_factor)
                )
                kickstart_multi = (
                    "0" if job.kickstart_multi is None else float(job.kickstart_multi)
                )
                remote_cpu_time = (
                    "0" if job.remote_cpu_time is None else float(job.remote_cpu_time)
                )
                post_time = "0" if job.post_time is None else float(job.post_time)
                condor_q_time = (
                    "0" if job.condor_q_time is None else float(job.condor_q_time)
                )
                resource_delay = (
                    "0" if job.resource_delay is None else float(job.resource_delay)
                )
                runtime = "0" if job.runtime is None else float(job.runtime)
                seqexec = "-" if job.seqexec is None else float(job.seqexec)

                seqexec_delay = "-"
                if job.seqexec is not None and job.kickstart is not None:
                    seqexec_delay = float(job.seqexec) - float(job.kickstart)

                if job.job_name in job_retry_count_dict:
                    job_retry_count_dict[job.job_name] += 1
                else:
                    job_retry_count_dict[job.job_name] = 1

                retry_count = job_retry_count_dict[job.job_name]

                content.append(
                    [
                        job.job_name,
                        retry_count,
                        job.site,
                        kickstart,
                        multiplier_factor,
                        kickstart_multi,
                        remote_cpu_time,
                        post_time,
                        condor_q_time,
                        resource_delay,
                        runtime,
                        seqexec,
                        seqexec_delay,
                        utils.raw_to_regular(job.exit_code),
                        job.host_name,
                    ]
                )

            return content

        finally:
            Dashboard.close(workflow)

    def integrity_stats(self):
        try:
            workflow = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url(), False
            )
            workflow.initialize(root_wf_id=self._wf_id)
            individual_stats = self._integrity_stats(workflow)

            workflow2 = stampede_statistics.StampedeStatistics(self.__get_wf_db_url())
            workflow2.initialize(self._root_wf_uuid)
            all_stats = self._integrity_stats(workflow2)

            return {"individual": individual_stats, "all": all_stats}

        finally:
            Dashboard.close(workflow)
            Dashboard.close(workflow2)

    def _integrity_stats(self, workflow):
        # tasks
        table = []
        workflow.set_job_filter("nonsub")
        stats = workflow.get_integrity_metrics()

        for s in stats:
            table.append(
                {
                    "type": s.type,
                    "file_type": s.file_type,
                    "count": int(s.count),
                    "duration": float(s.duration),
                }
            )

        return table

    def plots_gantt_chart(self):
        try:
            # Expand has to be set to false. The method does not provide information when expand set to True.
            workflow = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url(), False
            )
            workflow.initialize(root_wf_id=self._wf_id)
            gantt_chart = workflow.get_job_states()

            return gantt_chart
        finally:
            Dashboard.close(workflow)

    def plots_time_chart(self, wf_id, time_filter="hour"):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            details = workflow.get_workflow_information()

            workflow_plots = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url()
            )
            workflow_plots.initialize(details.wf_uuid)

            workflow_plots.set_job_filter("nonsub")
            workflow_plots.set_time_filter(time_filter)
            workflow_plots.set_transformation_filter(exclude=["condor::dagman"])

            job, invocation = (
                workflow_plots.get_jobs_run_by_time(),
                workflow_plots.get_invocation_by_time(),
            )

            return job, invocation

        finally:
            Dashboard.close(workflow)
            Dashboard.close(workflow_plots)

    def plots_transformation_statistics(self, wf_id):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            details = workflow.get_workflow_information()

            workflow_plots = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url()
            )
            workflow_plots.initialize(details.wf_uuid)

            workflow_plots.set_job_filter("nonsub")
            workflow_plots.set_time_filter("hour")
            workflow_plots.set_transformation_filter(exclude=["condor::dagman"])

            dist = workflow_plots.get_transformation_statistics()

            return dist

        finally:
            Dashboard.close(workflow)
            Dashboard.close(workflow_plots)

    def get_workflow_information(self, wf_id=None, wf_uuid=None):
        """
        Get workflow specific information. This is when user click on a workflow link.
        Returns a workflow object.
        """
        try:
            if not wf_id and not wf_uuid:
                raise ValueError("Workflow ID or Workflow UUID is required")

            workflow = None
            workflow_statistics = None
            workflow = queries.WorkflowInfo(
                self.__get_wf_db_url(), wf_id=wf_id, wf_uuid=wf_uuid
            )

            details = self._get_workflow_details(workflow)
            job_counts = self._get_workflow_job_counts(workflow)

            # workflow_statistics = stampede_statistics.StampedeStatistics(self.__get_wf_db_url(), expand_workflow=(details.root_wf_id ==  details.wf_id))
            workflow_statistics = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url(), expand_workflow=True
            )
            workflow_statistics.initialize(details.wf_uuid)

            statistics = {}

            statistics.update(self._get_workflow_summary_times(workflow_statistics))

            # if details.root_wf_id ==  details.wf_id:
            statistics.update(self._get_workflow_summary_counts(workflow_statistics))

            return job_counts, details, statistics

        finally:
            Dashboard.close(workflow)
            Dashboard.close(workflow_statistics)

    def get_workflow_details(self, wf_id=None, wf_uuid=None):
        """
        Get workflow specific information. This is when user click on a workflow link.
        Returns a workflow object.
        """
        try:
            if not wf_id and not wf_uuid:
                raise ValueError("Workflow ID or Workflow UUID is required")

            workflow = None
            workflow = queries.WorkflowInfo(
                self.__get_wf_db_url(), wf_id=wf_id, wf_uuid=wf_uuid
            )

            details = self._get_workflow_details(workflow)

            return details

        finally:
            Dashboard.close(workflow)

    def workflow_summary_stats(self, wf_id=None, wf_uuid=None):

        try:
            workflow = stampede_statistics.StampedeStatistics(
                self.__get_wf_db_url(), expand_workflow=False
            )
            workflow.initialize(root_wf_id=self._wf_id)
            dictionary = self._get_workflow_summary_times(workflow)
            dictionary["retry-count"] = self._get_workflow_retries(workflow)
            return dictionary
        finally:
            Dashboard.close(workflow)

    def _get_workflow_details(self, workflow):
        return workflow.get_workflow_information()

    def _get_workflow_job_counts(self, workflow):
        return workflow.get_workflow_job_counts()

    def _get_workflow_summary_times(self, workflow):
        statistics = {}

        workflow_states_list = workflow.get_workflow_states()

        wall_time = stats_utils.get_workflow_wall_time(workflow_states_list)
        if wall_time is not None:
            wall_time = float(wall_time)

        cum_time = workflow.get_workflow_cum_job_wall_time()
        cum_time = [float(v) if v is not None else v for v in cum_time]
        # if cum_time is not None:
        #    cum_time = float(cum_time)

        job_cum_time = workflow.get_submit_side_job_wall_time()
        job_cum_time = [float(v) if v is not None else v for v in job_cum_time]
        # if job_cum_time is not None:
        #    job_cum_time = float(job_cum_time)

        statistics["wall-time"] = wall_time
        statistics["cum-time"] = cum_time[0]
        statistics["cum-badput-time"] = cum_time[2]
        statistics["job-cum-time"] = job_cum_time[0]
        statistics["job-cum-badput-time"] = job_cum_time[2]

        return statistics

    def _get_workflow_retries(self, workflow):
        workflow.set_job_filter("all")
        return int(workflow.get_workflow_retries())

    def _get_workflow_summary_counts(self, workflow):
        statistics = {}

        workflow.set_job_filter("nonsub")
        statistics["total-jobs"] = workflow.get_total_jobs_status()

        statistics["successful-jobs"] = workflow.get_total_succeeded_jobs_status()
        statistics["failed-jobs"] = workflow.get_total_failed_jobs_status()
        statistics["running-jobs"] = workflow.get_total_running_jobs_status()
        statistics["unsubmitted-jobs"] = statistics["total-jobs"] - (
            statistics["successful-jobs"]
            + statistics["failed-jobs"]
            + statistics["running-jobs"]
        )
        statistics["job-retries"] = workflow.get_total_jobs_retries()
        statistics["job-instance-retries"] = (
            statistics["successful-jobs"]
            + statistics["failed-jobs"]
            + statistics["job-retries"]
        )

        return statistics

    def get_job_information(self, wf_id, job_id, job_instance_id):
        """
        Get job specific information. This is when user click on a job link, on the workflow details page.
        Returns a Job object.
        """
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            job_details = workflow.get_job_information(job_id, job_instance_id)
            return job_details
        except NoResultFound:
            return None
        finally:
            Dashboard.close(workflow)

    def get_job_instances(self, wf_id, job_id):
        """
        Get a list of all job instances for a given job
        """
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            job_instances = workflow.get_job_instances(job_id)
            return job_instances
        except NoResultFound:
            return None
        finally:
            Dashboard.close(workflow)

    def get_job_states(self, wf_id, job_id, job_instance_id):
        """
        Get information about the job states that a job has gone through.
        """
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            job_states = workflow.get_job_states(job_id, job_instance_id)
            return job_states
        finally:
            Dashboard.close(workflow)

    def get_failed_jobs(self, wf_id, **table_args):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id=wf_id)
            total_count, filtered_count, failed_jobs = workflow.get_failed_jobs(
                **table_args
            )
            return total_count, filtered_count, failed_jobs
        finally:
            Dashboard.close(workflow)

    def get_successful_jobs(self, wf_id, **table_args):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id=wf_id)
            total_count, filtered_count, successful_jobs = workflow.get_successful_jobs(
                **table_args
            )
            return total_count, filtered_count, successful_jobs
        finally:
            Dashboard.close(workflow)

    def get_running_jobs(self, wf_id, **table_args):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id=wf_id)
            total_count, filtered_count, running_jobs = workflow.get_other_jobs(
                **table_args
            )
            return total_count, filtered_count, running_jobs
        finally:
            Dashboard.close(workflow)

    def get_failing_jobs(self, wf_id, **table_args):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id=wf_id)
            total_count, filtered_count, failed_jobs = workflow.get_failing_jobs(
                **table_args
            )
            return total_count, filtered_count, failed_jobs
        finally:
            Dashboard.close(workflow)

    def get_sub_workflows(self, wf_id):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id=wf_id)
            sub_workflows = workflow.get_sub_workflows()
            return sub_workflows
        finally:
            Dashboard.close(workflow)

    def get_stdout(self, wf_id, job_id, job_instance_id):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            stdout = workflow.get_stdout(job_id, job_instance_id)
            return stdout
        finally:
            Dashboard.close(workflow)

    def get_successful_job_invocation(self, wf_id, job_id, job_instance_id):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            successful_invocations = workflow.get_successful_job_invocations(
                job_id, job_instance_id
            )
            return successful_invocations
        finally:
            Dashboard.close(workflow)

    def get_failed_job_invocation(self, wf_id, job_id, job_instance_id):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            failed_invocations = workflow.get_failed_job_invocations(
                job_id, job_instance_id
            )
            return failed_invocations
        finally:
            Dashboard.close(workflow)

    def get_stderr(self, wf_id, job_id, job_instance_id):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            stderr = workflow.get_stderr(job_id, job_instance_id)
            return stderr
        finally:
            Dashboard.close(workflow)

    def get_invocation_information(self, wf_id, job_id, job_instance_id, invocation_id):
        try:
            workflow = queries.WorkflowInfo(self.__get_wf_db_url(), wf_id)
            invocation = workflow.get_invocation_information(
                job_id, job_instance_id, invocation_id
            )
            return invocation
        finally:
            Dashboard.close(workflow)



#  Copyright 2007-2012 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__author__ = "Rajiv Mayani"

import logging
import time

from sqlalchemy import orm
from sqlalchemy.orm.exc import *
from sqlalchemy.sql.expression import (
    and_,
    asc,
    between,
    case,
    desc,
    distinct,
    func,
    or_,
)
from sqlalchemy.util._collections import KeyedTuple

from Pegasus.db import connection
from Pegasus.db.admin.admin_loader import DBAdminError
from Pegasus.db.errors import StampedeDBNotFoundError
from Pegasus.db.schema import *

log = logging.getLogger(__name__)


class MasterDBNotFoundError(Exception):
    pass


class MasterDatabase:
    def __init__(self, conn_string, debug=False):
        self._dbg = debug

        if conn_string is None:
            raise ValueError("Connection string is required")

        try:
            self.session = connection.connect(conn_string)
        except connection.ConnectionError as e:
            log.exception(e)
            message = e

            while isinstance(message, Exception):
                message = message.message

            if "attempt to write a readonly database" in message:
                raise DBAdminError(message)

            raise MasterDBNotFoundError(e)

    def close(self):
        log.debug("close")
        self.session.close()

    def get_wf_db_url(self, wf_id):
        """
        Given a work-flow UUID, query the master database to get the connection URL for the work-flow's STAMPEDE database.
        """

        w = orm.aliased(MasterWorkflow, name="w")

        q = self.session.query(w.db_url)
        q = q.filter(w.wf_id == wf_id)

        return q.one().db_url

    def get_wf_id_url(self, root_wf_id):
        """
        Given a work-flow UUID, query the master database to get the connection URL for the work-flow's STAMPEDE database.
        """
        q = self.session.query(MasterWorkflow)

        if root_wf_id is None:
            raise ValueError("root_wf_id cannot be None")

        m_wf_id = str(root_wf_id)
        if m_wf_id.isdigit():
            q = q.filter(MasterWorkflow.wf_id == root_wf_id)
        else:
            q = q.filter(MasterWorkflow.wf_uuid == root_wf_id)

        q = q.one()

        return q.wf_id, q.wf_uuid, q.db_url

    def get_all_workflows(self, **table_args):
        """
        SELECT w.*, ws.*
         FROM   workflow w
                     JOIN workflowstate ws ON w.wf_id = ws.wf_id
                JOIN (SELECT wf_id, max(timestamp) time
                            FROM   workflowstate
                            GROUP  BY wf_id) t ON ws.wf_id = t.wf_id
                           AND ws.timestamp = t.time
         WHERE  w.wf_id = ws.wf_id
        AND ws.wf_id = t.wf_id
        AND ws.timestamp = t.time;
        """

        w = orm.aliased(MasterWorkflow, name="w")
        ws = orm.aliased(MasterWorkflowstate, name="ws")

        # Get last state change for each work-flow.
        qmax = self.session.query(
            MasterWorkflowstate.wf_id,
            func.max(MasterWorkflowstate.timestamp).label("max_time"),
        )
        qmax = qmax.group_by(MasterWorkflowstate.wf_id)

        qmax = qmax.subquery("max_timestamp")

        state = case(
            [
                (ws.status == None, "Running"),  # noqa: E711
                (ws.status == 0, "Successful"),
                (ws.status != 0, "Failed"),
            ],
            else_="Undefined",
        ).label("state")

        q = self.session.query(
            w.wf_id,
            w.wf_uuid,
            w.timestamp,
            w.dag_file_name,
            w.submit_hostname,
            w.submit_dir,
            w.planner_arguments,
            w.user,
            w.grid_dn,
            w.planner_version,
            w.dax_label,
            w.dax_version,
            w.db_url,
            w.archived,
            ws.reason,
            ws.status,
            state,
        )

        q = q.filter(w.wf_id == ws.wf_id)
        q = q.filter(ws.wf_id == qmax.c.wf_id)
        q = q.filter(ws.timestamp == qmax.c.max_time)

        # Get Total Count. Need this to pass to jQuery Datatable.
        count = q.count()
        if count == 0:
            return 0, 0, []

        if "filter" in table_args:
            filter_text = "%" + table_args["filter"] + "%"
            q = q.filter(
                or_(
                    w.dax_label.like(filter_text),
                    w.submit_hostname.like(filter_text),
                    w.submit_dir.like(filter_text),
                    state.like(filter_text),
                )
            )

        if "time_filter" in table_args:
            time_filter = table_args["time_filter"]
            current_time = int(time.time())

            if time_filter == "day":
                q = q.filter(between(w.timestamp, current_time - 86400, current_time))
            elif time_filter == "week":
                q = q.filter(between(w.timestamp, current_time - 604800, current_time))
            elif time_filter == "month":
                q = q.filter(between(w.timestamp, current_time - 2620800, current_time))
            elif time_filter == "year":
                q = q.filter(
                    between(w.timestamp, current_time - 31449600, current_time)
                )

        # Get Total Count. Need this to pass to jQuery Datatable.
        filtered = q.count()

        if filtered == 0:
            return count, 0, []

        display_columns = [
            w.dax_label,
            w.submit_hostname,
            w.submit_dir,
            state,
            w.timestamp,
        ]

        if "sort-col-count" in table_args:
            for i in range(table_args["sort-col-count"]):

                if "iSortCol_" + str(i) in table_args:
                    if (
                        "sSortDir_" + str(i) in table_args
                        and table_args["sSortDir_" + str(i)] == "asc"
                    ):
                        i = table_args["iSortCol_" + str(i)]

                        if 0 <= i < len(display_columns):
                            q = q.order_by(display_columns[i])
                        else:
                            raise ValueError(
                                "Invalid column (%s) in work-flow listing " % i
                            )
                    else:
                        i = table_args["iSortCol_" + str(i)]

                        if 0 <= i < len(display_columns):
                            q = q.order_by(desc(display_columns[i]))
                        else:
                            raise ValueError(
                                "Invalid column (%s) in work-flow listing " % i
                            )

        else:
            # Default sorting order
            q = q.order_by(desc(w.timestamp))

        if "limit" in table_args and "offset" in table_args:
            q = q.limit(table_args["limit"])
            q = q.offset(table_args["offset"])

        return count, filtered, q.all()

    def get_workflow_counts(self):

        w = orm.aliased(MasterWorkflow, name="w")
        ws = orm.aliased(MasterWorkflowstate, name="ws")

        # Get last state change for each work-flow.
        qmax = self.session.query(
            MasterWorkflowstate.wf_id,
            func.max(MasterWorkflowstate.timestamp).label("max_time"),
        )
        qmax = qmax.group_by(MasterWorkflowstate.wf_id)

        qmax = qmax.subquery("max_timestamp")

        q = self.session.query(
            func.count(w.wf_id).label("total"),
            func.sum(case([(ws.status == 0, 1)], else_=0)).label("success"),
            func.sum(case([(ws.status != 0, 1)], else_=0)).label("fail"),
            func.sum(case([(ws.status == None, 1)], else_=0)).label(
                "others"
            ),  # noqa: E711
        )

        q = q.filter(w.wf_id == ws.wf_id)
        q = q.filter(ws.wf_id == qmax.c.wf_id)
        q = q.filter(ws.timestamp == qmax.c.max_time)

        return q.one()


class WorkflowInfo:
    def __init__(self, conn_string=None, wf_id=None, wf_uuid=None, debug=False):
        self._dbg = debug

        if conn_string is None:
            raise ValueError("Connection string is required")

        try:
            self.session = connection.connect(conn_string)
        except connection.ConnectionError as e:
            log.exception(e)
            message = e

            while isinstance(message, Exception):
                message = message.message

            if "attempt to write a readonly database" in message:
                raise DBAdminError(message)

            raise StampedeDBNotFoundError

        self.initialize(wf_id, wf_uuid)

    def initialize(self, wf_id=None, wf_uuid=None):

        if not wf_id and not wf_uuid:
            raise ValueError("Workflow ID or Workflow UUID is required.")

        if wf_id:
            self._wf_id = wf_id
            return

        if wf_uuid:
            q = self.session.query(Workflow.wf_id)
            q = q.filter(Workflow.wf_uuid == wf_uuid)

            self._wf_id = q.one().wf_id

    def get_workflow_information(self):

        qmax = self.session.query(func.max(Workflowstate.timestamp).label("max_time"))
        qmax = qmax.filter(Workflowstate.wf_id == self._wf_id)

        qmax = qmax.subquery("max_timestamp")

        ws = orm.aliased(Workflowstate, name="ws")
        w = orm.aliased(Workflow, name="w")

        q = self.session.query(
            w.wf_id,
            w.wf_uuid,
            w.parent_wf_id,
            w.root_wf_id,
            w.dag_file_name,
            w.submit_hostname,
            w.submit_dir,
            w.planner_arguments,
            w.user,
            w.grid_dn,
            w.planner_version,
            w.dax_label,
            w.dax_version,
            case(
                [
                    (ws.status == None, "Running"),  # noqa: E711
                    (ws.status == 0, "Successful"),
                    (ws.status != 0, "Failed"),
                ],
                else_="Undefined",
            ).label("state"),
            ws.reason,
            ws.timestamp,
        )

        q = q.filter(w.wf_id == self._wf_id)
        q = q.filter(w.wf_id == ws.wf_id)
        q = q.filter(ws.timestamp == qmax.c.max_time)

        return q.one()

    def get_workflow_job_counts(self):

        qmax = self.__get_maxjss_subquery()

        q = self.session.query(func.count(Job.wf_id).label("total"))
        q = q.add_column(
            func.sum(
                case(
                    [(Job.type_desc == "dag", 1), (Job.type_desc == "dax", 1)], else_=0
                )
            ).label("total_workflow")
        )
        q = q.filter(Job.wf_id == self._wf_id)

        totals = q.one()

        q = self.session.query(func.count(Job.wf_id).label("total"))
        q = q.add_column(
            func.sum(case([(JobInstance.exitcode == 0, 1)], else_=0)).label("success")
        )
        q = q.add_column(
            func.sum(
                case(
                    [
                        (
                            JobInstance.exitcode == 0,
                            case(
                                [
                                    (Job.type_desc == "dag", 1),
                                    (Job.type_desc == "dax", 1),
                                ],
                                else_=0,
                            ),
                        )
                    ],
                    else_=0,
                )
            ).label("success_workflow")
        )

        q = q.add_column(
            func.sum(case([(JobInstance.exitcode != 0, 1)], else_=0)).label("fail")
        )
        q = q.add_column(
            func.sum(
                case(
                    [
                        (
                            JobInstance.exitcode != 0,
                            case(
                                [
                                    (Job.type_desc == "dag", 1),
                                    (Job.type_desc == "dax", 1),
                                ],
                                else_=0,
                            ),
                        )
                    ],
                    else_=0,
                )
            ).label("fail_workflow")
        )

        q = q.add_column(
            func.sum(
                case([(JobInstance.exitcode == None, 1)], else_=0)
            ).label(  # noqa: E711
                "running"
            )
        )
        q = q.add_column(
            func.sum(
                case(
                    [
                        (
                            JobInstance.exitcode == None,  # noqa: E711
                            case(
                                [
                                    (Job.type_desc == "dag", 1),
                                    (Job.type_desc == "dax", 1),
                                ],
                                else_=0,
                            ),
                        )
                    ],
                    else_=0,
                )
            ).label("running_workflow")
        )

        q = q.filter(Job.wf_id == self._wf_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(Job.job_id == qmax.c.job_id)
        q = q.filter(JobInstance.job_submit_seq == qmax.c.max_jss)

        counts = q.one()

        out = KeyedTuple(
            [
                totals.total,
                totals.total_workflow,
                totals.total - (counts.success + counts.fail + counts.running),
                totals.total_workflow
                - (
                    counts.success_workflow
                    + counts.fail_workflow
                    + counts.running_workflow
                ),
                counts.success,
                counts.success_workflow,
                counts.fail,
                counts.fail_workflow,
                counts.running,
                counts.running_workflow,
            ],
            labels=[
                "total",
                "total_workflow",
                "others",
                "others_workflow",
                "success",
                "success_workflow",
                "fail",
                "fail_workflow",
                "running",
                "running_workflow",
            ],
        )

        return out

    def get_job_information(self, job_id, job_instance_id):

        q = self.session.query(
            Job.exec_job_id,
            Job.clustered,
            JobInstance.job_instance_id,
            JobInstance.work_dir,
            JobInstance.exitcode,
            JobInstance.stdout_file,
            JobInstance.stderr_file,
            Host.site,
            Host.hostname,
            Host.ip,
        )
        q = q.filter(Job.wf_id == self._wf_id)
        q = q.filter(Job.job_id == job_id)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.outerjoin(Host, JobInstance.host_id == Host.host_id)

        return q.one()

    def get_job_instances(self, job_id):

        q = self.session.query(
            Job.exec_job_id,
            JobInstance.job_instance_id,
            JobInstance.exitcode,
            JobInstance.job_submit_seq,
        )
        q = q.filter(Job.wf_id == self._wf_id)
        q = q.filter(Job.job_id == job_id)
        q = q.filter(Job.job_id == JobInstance.job_id)

        q = q.order_by(desc(JobInstance.job_submit_seq))

        return q.all()

    def get_job_states(self, job_id, job_instance_id):

        q = self.session.query(Jobstate.state, Jobstate.reason, Jobstate.timestamp)
        q = q.filter(Job.wf_id == self._wf_id)
        q = q.filter(Job.job_id == job_id)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(JobInstance.job_instance_id == Jobstate.job_instance_id)

        q = q.order_by(asc(Jobstate.jobstate_submit_seq))

        return q.all()

    def _jobs_by_type(self):
        qmax = self.__get_jobs_maxjss_sq()

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            Job.exec_job_id,
            JobInstance.exitcode,
        )

        q = q.filter(Job.wf_id == self._wf_id)

        q = q.filter(Job.job_id == JobInstance.job_id)

        q = q.filter(Job.job_id == qmax.c.job_id)
        q = q.filter(JobInstance.job_submit_seq == qmax.c.max_jss)

        q = q.group_by(JobInstance.job_id)

        return q

    def get_failed_jobs(self, **table_args):

        q = self._jobs_by_type()
        q = q.filter(JobInstance.exitcode != 0).filter(
            JobInstance.exitcode != None
        )  # noqa: E711

        # Get Total Count. Need this to pass to jQuery Datatable.
        count = q.count()
        if count == 0:
            return (0, 0, [])

        filtered = count
        if "filter" in table_args:
            filter_text = "%" + table_args["filter"] + "%"
            q = q.filter(
                or_(
                    Job.exec_job_id.like(filter_text),
                    JobInstance.exitcode.like(filter_text),
                )
            )

            # Get Total Count. Need this to pass to jQuery Datatable.
            filtered = q.count()

            if filtered == 0:
                return (count, 0, [])

        display_columns = [Job.exec_job_id, JobInstance.exitcode]

        if "sort-col-count" in table_args:
            for i in range(table_args["sort-col-count"]):

                if "iSortCol_" + str(i) in table_args:
                    sort_order = desc

                    if (
                        "sSortDir_" + str(i) in table_args
                        and table_args["sSortDir_" + str(i)] == "asc"
                    ):
                        sort_order = asc

                    i = table_args["iSortCol_" + str(i)]

                    if i >= 0 and i < len(display_columns):
                        q = q.order_by(sort_order(display_columns[i]))
                    elif i >= len(display_columns) and i < 4:
                        pass
                    else:
                        raise ValueError(
                            "Invalid column(%s) in failed jobs listing " % i
                        )

        else:
            # Default sorting order
            q = q.order_by(desc(Job.exec_job_id))

        if "limit" in table_args and "offset" in table_args:
            q = q.limit(table_args["limit"])
            q = q.offset(table_args["offset"])

        return count, filtered, q.all()

    def get_successful_jobs(self, **table_args):

        q = self._jobs_by_type()

        q = q.add_column(JobInstance.local_duration)
        q = q.add_column(JobInstance.cluster_duration)
        duration = case(
            [(Job.clustered == 1, JobInstance.cluster_duration)],
            else_=JobInstance.local_duration,
        ).label("duration")
        q = q.add_column(duration)

        q = q.filter(JobInstance.exitcode == 0).filter(
            JobInstance.exitcode != None
        )  # noqa: E711

        # Get Total Count. Need this to pass to jQuery Datatable.
        count = q.count()
        if count == 0:
            return (0, 0, [])

        filtered = count
        if "filter" in table_args:
            filter_text = "%" + table_args["filter"] + "%"
            q = q.filter(or_(Job.exec_job_id.like(filter_text)))

            # Get Total Count. Need this to pass to jQuery Datatable.
            filtered = q.count()

            if filtered == 0:
                return count, 0, []

        display_columns = [Job.exec_job_id, duration]

        if "sort-col-count" in table_args:
            for i in range(table_args["sort-col-count"]):

                if "iSortCol_" + str(i) in table_args:
                    sort_order = desc

                    if (
                        "sSortDir_" + str(i) in table_args
                        and table_args["sSortDir_" + str(i)] == "asc"
                    ):
                        sort_order = asc

                    i = table_args["iSortCol_" + str(i)]

                    if i >= 0 and i < len(display_columns):
                        q = q.order_by(sort_order(display_columns[i]))
                    else:
                        raise ValueError(
                            "Invalid column(%s) in successful jobs listing " % i
                        )

        else:
            # Default sorting order
            q = q.order_by(desc(Job.exec_job_id))

        if "limit" in table_args and "offset" in table_args:
            q = q.limit(table_args["limit"])
            q = q.offset(table_args["offset"])

        return count, filtered, q.all()

    def get_other_jobs(self, **table_args):

        q = self._jobs_by_type()

        q = q.add_column(JobInstance.local_duration)
        q = q.add_column(JobInstance.cluster_duration)
        q = q.add_column(
            case(
                [(Job.clustered == 1, JobInstance.cluster_duration)],
                else_=JobInstance.local_duration,
            ).label("duration")
        )

        q = q.filter(JobInstance.exitcode == None)  # noqa: E711

        # Get Total Count. Need this to pass to jQuery Datatable.
        count = q.count()
        if count == 0:
            return 0, 0, []

        filtered = count
        if "filter" in table_args:
            filter_text = "%" + table_args["filter"] + "%"
            q = q.filter(or_(Job.exec_job_id.like(filter_text)))

            # Get Total Count. Need this to pass to jQuery Datatable.
            filtered = q.count()

            if filtered == 0:
                return count, 0, []

        display_columns = [Job.exec_job_id]

        if "sort-col-count" in table_args:
            for i in range(table_args["sort-col-count"]):

                if "iSortCol_" + str(i) in table_args:
                    sort_order = desc

                    if (
                        "sSortDir_" + str(i) in table_args
                        and table_args["sSortDir_" + str(i)] == "asc"
                    ):
                        sort_order = asc

                    i = table_args["iSortCol_" + str(i)]

                    if i >= 0 and i < len(display_columns):
                        q = q.order_by(sort_order(display_columns[i]))
                    else:
                        raise ValueError(
                            "Invalid column(%s) in other jobs listing " % i
                        )

        else:
            # Default sorting order
            q = q.order_by(desc(Job.exec_job_id))

        if "limit" in table_args and "offset" in table_args:
            q = q.limit(table_args["limit"])
            q = q.offset(table_args["offset"])

        return count, filtered, q.all()

    def get_failing_jobs(self, **table_args):
        """
        SELECT job.job_id                   AS job_job_id,
               job_instance.job_instance_id AS job_instance_job_instance_id,
               job.exec_job_id              AS job_exec_job_id,
               job_instance.exitcode        AS job_instance_exitcode
        FROM   job,
               job_instance,
               (SELECT job.job_id                       AS job_id,
                       Max(job_instance.job_submit_seq) AS max_jss
                FROM   job,
                       job_instance
                WHERE  job.wf_id = 1
                       AND job.job_id = job_instance.job_id
                       AND job_instance.exitcode IS NOT NULL
                       AND job_instance.exitcode != 0
                GROUP  BY job.job_id) AS allmaxjss
        WHERE  job.wf_id = 1
               AND job.job_id = job_instance.job_id
               AND job_instance.exitcode != 0
               AND job_instance.exitcode IS NOT NULL
               AND job.job_id = allmaxjss.job_id
               AND job_instance.job_submit_seq = allmaxjss.max_jss
               AND job.job_id IN (SELECT DISTINCT j1.job_id AS anon_1
                                  FROM   job AS j1,
                                         job_instance AS ji1
                                  WHERE  j1.wf_id = 1
                                         AND j1.job_id = ji1.job_id
                                         AND ji1.exitcode IS NULL);
        """
        # Get a list of running jobs.
        j1 = orm.aliased(Job, name="j1")
        ji1 = orm.aliased(JobInstance, name="ji1")

        q_sub = self.session.query(distinct(j1.job_id))

        q_sub = q_sub.filter(j1.wf_id == self._wf_id)

        q_sub = q_sub.filter(j1.job_id == ji1.job_id)

        q_sub = q_sub.filter(ji1.exitcode == None)  # noqa: E711

        q_sub = q_sub.subquery()

        #
        # Get max(job_submit_seq) of all the failed job instances, for each job.
        #
        qmax = self.__get_jobs_maxjss_q()
        qmax = qmax.filter(JobInstance.exitcode != None).filter(
            JobInstance.exitcode != 0
        )  # noqa: E711
        qmax = qmax.subquery("allmaxjss")

        #
        # Get the latest failed job instances
        # whose job_id matches the job_ids of the currently running jobs.
        #

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            Job.exec_job_id,
            JobInstance.exitcode,
        )

        q = q.filter(Job.wf_id == self._wf_id)

        q = q.filter(Job.job_id == JobInstance.job_id)

        q = q.filter(JobInstance.exitcode != 0).filter(
            JobInstance.exitcode != None
        )  # noqa: E711

        q = q.filter(Job.job_id == qmax.c.job_id)
        q = q.filter(JobInstance.job_submit_seq == qmax.c.max_jss)

        q = q.filter(Job.job_id.in_(q_sub))

        #
        # Get Total Count. Need this to pass to jQuery Datatable.
        #
        count = q.count()
        if count == 0:
            return 0, 0, []

        filtered = count
        if "filter" in table_args:
            filter_text = "%" + table_args["filter"] + "%"
            q = q.filter(
                or_(
                    Job.exec_job_id.like(filter_text),
                    JobInstance.exitcode.like(filter_text),
                )
            )

            # Get Total Count. Need this to pass to jQuery Datatable.
            filtered = q.count()

            if filtered == 0:
                return count, 0, []

        display_columns = [Job.exec_job_id, JobInstance.exitcode]

        if "sort-col-count" in table_args:
            for i in range(table_args["sort-col-count"]):

                if "iSortCol_" + str(i) in table_args:
                    sort_order = desc

                    if (
                        "sSortDir_" + str(i) in table_args
                        and table_args["sSortDir_" + str(i)] == "asc"
                    ):
                        sort_order = asc

                    i = table_args["iSortCol_" + str(i)]

                    if i >= 0 and i < len(display_columns):
                        q = q.order_by(sort_order(display_columns[i]))
                    elif i >= len(display_columns) and i < 4:
                        pass
                    else:
                        raise ValueError(
                            "Invalid column(%s) in failed jobs listing " % i
                        )

        else:
            # Default sorting order
            q = q.order_by(desc(Job.exec_job_id))

        if "limit" in table_args and "offset" in table_args:
            q = q.limit(table_args["limit"])
            q = q.offset(table_args["offset"])

        return count, filtered, q.all()

    def __get_jobs_maxjss_q(self):
        qmax = self.session.query(
            Job.job_id, func.max(JobInstance.job_submit_seq).label("max_jss")
        )
        qmax = qmax.filter(Job.wf_id == self._wf_id)
        qmax = qmax.filter(Job.job_id == JobInstance.job_id)
        qmax = qmax.group_by(Job.job_id)

        return qmax

    def __get_jobs_maxjss_sq(self):
        qmax = self.__get_jobs_maxjss_q()
        qmax = qmax.subquery("allmaxjss")
        return qmax

    def get_sub_workflows(self):
        qmax = self.session.query(
            Workflowstate.wf_id, func.max(Workflowstate.timestamp).label("max_time")
        )
        qmax = qmax.group_by(Workflowstate.wf_id)

        qmax = qmax.subquery("max_timestamp")

        ws = orm.aliased(Workflowstate, name="ws")
        w = orm.aliased(Workflow, name="w")

        q = self.session.query(
            w.root_wf_id,
            w.wf_id,
            w.wf_uuid,
            w.dax_label,
            case(
                [
                    (ws.status == None, "Running"),  # noqa: E711
                    (ws.status == 0, "Successful"),
                    (ws.status != 0, "Failed"),
                ],
                else_="Undefined",
            ).label("state"),
        )

        q = q.filter(w.parent_wf_id == self._wf_id)
        q = q.filter(w.wf_id == ws.wf_id)
        q = q.filter(ws.wf_id == qmax.c.wf_id)
        q = q.filter(ws.timestamp == qmax.c.max_time)

        return q.all()

    def get_stdout(self, job_id, job_instance_id):
        q = self.session.query(JobInstance.stdout_file, JobInstance.stdout_text)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)

        return q.one()

    def get_stderr(self, job_id, job_instance_id):
        q = self.session.query(JobInstance.stderr_file, JobInstance.stderr_text)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)

        return q.one()

    def get_successful_job_invocations(self, job_id, job_instance_id):

        q = self.session.query(
            Job.exec_job_id,
            Invocation.invocation_id,
            Invocation.abs_task_id,
            Invocation.exitcode,
            Invocation.remote_duration,
        )
        q = q.filter(Job.wf_id == self._wf_id)
        q = q.filter(Job.job_id == job_id)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(JobInstance.job_instance_id == Invocation.job_instance_id)
        q = q.filter(Invocation.exitcode == 0)

        q = q.filter(
            or_(
                Invocation.abs_task_id != None, Invocation.task_submit_seq == 1
            )  # noqa: E711
        )

        return q.all()

    def get_failed_job_invocations(self, job_id, job_instance_id):

        q = self.session.query(
            Job.exec_job_id,
            Invocation.invocation_id,
            Invocation.abs_task_id,
            Invocation.exitcode,
            Invocation.remote_duration,
        )
        q = q.filter(Job.wf_id == self._wf_id)
        q = q.filter(Job.job_id == job_id)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(JobInstance.job_instance_id == Invocation.job_instance_id)
        q = q.filter(Invocation.exitcode != 0)

        q = q.filter(
            or_(
                Invocation.abs_task_id != None, Invocation.task_submit_seq == 1
            )  # noqa: E711
        )

        return q.all()

    def __get_maxjss_subquery(self, job_id=None):

        jii = orm.aliased(JobInstance, name="jii")

        if job_id:

            qmax = self.session.query(
                JobInstance.job_instance_id, func.max(JobInstance.job_submit_seq)
            )
            qmax = qmax.filter(Job.wf_id == self._wf_id)
            qmax = qmax.filter(Job.job_id == job_id)
            qmax = qmax.filter(Job.type_desc != "dax", Job.type_desc != "dag")
            qmax = qmax.filter(Job.job_id == JobInstance.job_id).correlate(jii)

            qmax = qmax.subquery("maxjss")

        else:

            qmax = self.session.query(
                Job.job_id, func.max(JobInstance.job_submit_seq).label("max_jss")
            )
            qmax = qmax.filter(Job.wf_id == self._wf_id)
            qmax = qmax.filter(Job.job_id == JobInstance.job_id).correlate(jii)

            qmax = qmax.group_by(Job.job_id)

            qmax = qmax.subquery("maxjss")

        return qmax

    def get_invocation_information(self, job_id, job_instance_id, invocation_id):

        q = self.session.query(JobInstance.work_dir)

        q = q.join(Job, Job.job_id == JobInstance.job_id)
        q = q.outerjoin(Task, Job.job_id == Task.job_id)
        q = q.join(
            Invocation,
            and_(
                JobInstance.job_instance_id == Invocation.job_instance_id,
                and_(
                    or_(
                        Task.abs_task_id == None,  # noqa: E711
                        and_(
                            Task.abs_task_id != None,  # noqa: E711
                            Task.abs_task_id == Invocation.abs_task_id,
                        ),
                    )
                ),
            ),
        )

        q = q.filter(Job.wf_id == self._wf_id)
        q = q.filter(Job.job_id == job_id)
        q = q.filter(JobInstance.job_instance_id == job_instance_id)

        q = q.add_columns(
            Task.task_id,
            Invocation.invocation_id,
            Invocation.abs_task_id,
            Invocation.start_time,
            Invocation.remote_duration,
            Invocation.remote_cpu_time,
            Invocation.exitcode,
            Invocation.transformation,
            Invocation.executable,
            Invocation.argv,
        )

        if invocation_id is None:
            q = q.filter(Invocation.task_submit_seq == 1)
        else:
            q = q.filter(Invocation.invocation_id == invocation_id)

        return q.one()

    def close(self):
        log.debug("close")
        self.session.close()



__author__ = "Monte Goode"
__author__ = "Karan Vahi"

import time

from sqlalchemy import exc, orm

from Pegasus.db.base_loader import BaseLoader
from Pegasus.db.schema import *
from Pegasus.netlogger import util


class WorkflowLoader(BaseLoader):
    """Load into the Stampede SQL schema through SQLAlchemy.

    Parameters:
      - connString {string,None*}: SQLAlchemy connection string.
        The general form of this is
          'dialect+driver://username:password@host:port/database'.
        See the SQLAlchemy docs for details.
        For sqlite, use 'sqlite:///foo.db' for a relative path and
        'sqlite:////path/to/foo.db' (four slashes) for an absolute one.
        When using MySQL, the general form will work, but the library
        expects the database to exist (ie: will not issue CREATE DB)
        but will populate an empty DB with tables/indexes/etc.
    """

    MAX_RETRIES = 10  # maximum number of retries in case of operational errors that arise because of database locked/connection dropped

    def __init__(
        self,
        connString,
        perf=False,
        batch=False,
        props=None,
        db_type=None,
        backup=False,
    ):
        """Init object

        @type   connString: string
        @param  connString: SQLAlchemy connection string - REQUIRED
        """
        super().__init__(
            connString,
            batch=batch,
            props=props,
            db_type=db_type,
            backup=backup,
            flush_every=1000,
        )

        # "Case" dict to map events to handler methods
        self.eventMap = {
            "stampede.wf.plan": self.workflow,
            "stampede.wf.map.task_job": self.task_map,
            "stampede.static.start": self.noop,  # good
            "stampede.static.end": self.static_end,
            "stampede.xwf.start": self.workflowstate,
            "stampede.xwf.end": self.workflowstate,
            "stampede.xwf.map.subwf_job": self.subwf_map,
            "stampede.task.info": self.task,
            "stampede.task.edge": self.task_edge,
            "stampede.job.info": self.job,
            "stampede.job.edge": self.job_edge,
            "stampede.job_inst.pre.start": self.job_instance,
            "stampede.job_inst.pre.term": self.jobstate,
            "stampede.job_inst.pre.end": self.jobstate,
            "stampede.job_inst.submit.start": self.job_instance,
            "stampede.job_inst.submit.end": self.jobstate,
            "stampede.job_inst.held.start": self.jobstate,
            "stampede.job_inst.held.end": self.jobstate,
            "stampede.job_inst.main.start": self.jobstate,
            "stampede.job_inst.main.term": self.jobstate,
            "stampede.job_inst.main.end": self.job_instance,
            "stampede.job_inst.post.start": self.jobstate,
            "stampede.job_inst.post.term": self.jobstate,
            "stampede.job_inst.post.end": self.job_instance,
            "stampede.job_inst.host.info": self.host,
            "stampede.job_inst.image.info": self.jobstate,
            "stampede.job_inst.abort.info": self.jobstate,
            "stampede.job_inst.grid.submit.start": self.noop,  # good
            "stampede.job_inst.grid.submit.end": self.jobstate,
            "stampede.job_inst.globus.submit.start": self.noop,  # good
            "stampede.job_inst.globus.submit.end": self.jobstate,
            "stampede.job_inst.tag": self.tag,
            "stampede.job_inst.composite": self.noop,
            "stampede.inv.start": self.noop,  # good
            "stampede.inv.end": self.invocation,
            "stampede.static.meta.start": self.static_meta_start,
            "stampede.xwf.meta": self.workflow_meta,
            "stampede.task.meta": self.task_meta,
            "stampede.rc.meta": self.rc_meta,
            "stampede.int.metric": self.int_metric,
            "stampede.rc.pfn": self.rc_pfn,
            "stampede.wf.map.file": self.wf_task_file_map,
            "stampede.static.meta.end": self.noop,
            "stampede.task.monitoring": self.noop,
        }

        # Dicts for caching FK lookups
        self.wf_id_cache = {}
        self.root_wf_id_cache = {}
        self.task_id_cache = {}  # for task metadata population
        self.lfn_id_cache = {}  # for file metadata population
        self.job_id_cache = {}
        self.job_instance_id_cache = {}
        self.host_cache = {}
        self.hosts_written_cache = None

        # undocumented performance option
        self._perf = perf
        if self._perf:
            self._insert_time, self._insert_num = 0, 0
            self._start_time = time.time()

        # caches for batched events
        self._batch_cache = {
            "batch_events": [],
            "update_events": [],
            "host_map_events": [],
        }
        self._task_map_flush = {}
        self._task_edge_flush = {}

    def process(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Get the BP dict from the controlling process and dispatch
        to the appropriate method per-event.
        """
        self.log.trace("Process: %s", linedata)

        for retry in range(1, self.MAX_RETRIES + 1):
            if not self._batch:
                self.check_connection()

            try:
                if self._perf:
                    t = time.time()
                    self.eventMap[linedata["event"]](linedata)
                    self._insert_time += time.time() - t
                    self._insert_num += 1
                else:
                    self.eventMap[linedata["event"]](linedata)

            except KeyError:
                if linedata["event"].startswith("stampede.job_inst."):
                    self.log.warning(
                        'Corner case jobstate event: "%s"', linedata["event"]
                    )
                    self.jobstate(linedata)
                else:
                    self.log.error(
                        'No handler for event type "%s" defined', linedata["event"]
                    )
            except exc.IntegrityError as e:
                # This is raised when an attempted insert violates the
                # schema (unique indexes, etc).
                self.log.exception(e)
                self.log.error('Insert failed for event "%s"', linedata["event"])
                self.session.rollback()
            except exc.OperationalError as e:
                self.log.error(
                    "Connection seemingly lost - attempting to refresh. Retry %s"
                    % retry
                )
                self.session.rollback()
                self.check_connection()
                # PM-1013 retry only in case of operational errors
                continue
            # the current attempt was successful or there was integrity error/key error. exit the loop
            break
        else:
            # loop finished after all retries have been made
            self.log.error(
                "Maximum number of retries reached for stampede_loader.process() method %s"
                % retry
            )
            raise RuntimeError(
                "Maximum number of retries reached for stampede_loader.process() method %s"
                % retry
            )

        self.check_flush(increment=True)

    def linedataToObject(self, linedata, o):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.
        @type   o: instance of mapper class from stampede_schema module.
        @param  o: Passed in by the appropriate event handler method.

        Takes the dict of BP linedata, assigns contents to the class o
        as attributes, and does any global type massaging like
        transforming dict strings to numeric types.
        """
        for k, v in linedata.items():
            if k == "level":
                continue

            # undot
            attr = k.replace(".", "_")

            attr_remap = {
                # workflow
                "xwf_id": "wf_uuid",
                "parent_xwf_id": "parent_wf_id",
                # task.info
                "task_id": "abs_task_id",
                # task.edge
                "child_task_id": "child_abs_task_id",
                "parent_task_id": "parent_abs_task_id",
                # job.info
                "job_id": "exec_job_id",
                # job.edge
                "child_job_id": "child_exec_job_id",
                "parent_job_id": "parent_exec_job_id",
                # xwf.start/end (none)
                # job_inst.submit.start/job_inst.submit.start/etc
                "job_inst_id": "job_submit_seq",
                "js_id": "jobstate_submit_seq",
                "cluster_dur": "cluster_duration",
                "local_dur": "local_duration",
                # inv.end
                "inv_id": "task_submit_seq",
                "dur": "remote_duration",
            }

            # remap attr names
            if attr in attr_remap:
                attr = attr_remap[attr]

            # sanitize argv input
            if attr == "argv":
                if v is not None:
                    v = v.replace("\\", "\\\\")
                    v = v.replace("'", "\\'")

            try:
                setattr(o, attr, v)
            except Exception:
                self.log.error("Unable to process attribute %s with values: %s", k, v)

        # global type re-assignments
        if hasattr(o, "ts"):
            # make all timestamp values floats
            o.ts = float(o.ts)
        if hasattr(o, "start_time") and o.start_time is not None:
            o.start_time = float(o.start_time)
        if hasattr(o, "cluster_start_time") and o.cluster_start_time is not None:
            o.cluster_start_time = float(o.cluster_start_time)
        if hasattr(o, "duration") and o.duration is not None:
            o.duration = float(o.duration)
        if hasattr(o, "restart_count") and o.restart_count is not None:
            o.restart_count = int(o.restart_count)
        return o

    #############################################
    # Methods to handle batching/flushing
    #############################################

    def hard_flush(self, batch_flush=True, retry=0):
        """
        @type   batch_flush: boolean
        @param  batch_flush: Defaults to true.  Is set to false
            when the batch commit hits and integrity error.

        Process queued inserts and flush/commit to the database.
        If the commit fails due to an integrity error, then method
        re-calls itself with setting batch_flush to False which
        causes each insert/object to be committed individually
        so all the "good" inserts can succeed.  This will increase
        the processing time of the batch with the bad data in it.
        """
        if not self._batch:
            return

        self.log.debug("Hard flush: batch_flush=%s", batch_flush)

        if retry == self.MAX_RETRIES + 1:
            # PM-1013 see if max retries is reached
            self.log.error(
                "Maximum number of retries reached for workflow_loader.hard_flush() method %s"
                % self.MAX_RETRIES
            )
            raise RuntimeError(
                "Maximum number of retries reached for workflow_loader.hard_flush() method %s"
                % self.MAX_RETRIES
            )

        retry = retry + 1
        self.check_connection()

        if self._perf:
            s = time.time()

        end_event = []

        self.log.debug(
            "Batch event sizes: batch_event_size=%s update_event_size=%s",
            len(self._batch_cache["batch_events"]),
            len(self._batch_cache["update_events"]),
        )

        for event in self._batch_cache["batch_events"]:
            if event.event == "stampede.xwf.end":
                end_event.append(event)
            if batch_flush:
                self.session.add(event)
            else:
                self.individual_commit(event)

        for event in self._batch_cache["update_events"]:
            if batch_flush:
                self.session.merge(event)
            else:
                self.individual_commit(event, merge=True)

        try:
            self.session.commit()
        except exc.IntegrityError as e:
            self.log.exception(e)
            self.log.error(
                "Integrity error on batch flush: batch will need to be committed per-event which will take longer"
            )
            self.session.rollback()
            self.hard_flush(batch_flush=False, retry=retry)
        except exc.OperationalError as e:
            self.log.exception(e)
            self.log.error(
                "Connection problem during commit in hard_flush(): reattempting batch. Retry %s"
                % retry
            )
            self.session.rollback()
            self.hard_flush(retry=retry)

        for host in self._batch_cache["host_map_events"]:
            self.map_host_to_job_instance(host)

        for ee in end_event:
            self.purgeCaches(ee)
        end_event = []

        # Clear all data structures here.
        for k in self._batch_cache.keys():
            self._batch_cache[k] = []

        try:
            # commit the map host to job events . no retries for this.
            self.session.commit()
        except exc.IntegrityError as e:
            self.log.exception(e)
            self.log.error("Integrity error on host_map_events in hard_flush()")
            self.session.rollback()
        except exc.OperationalError as e:
            self.log.exception(e)
            self.log.error(
                "Connection problem on host_map_events during commit in hard_flush()"
            )
            self.session.rollback()

        self.reset_flush_state()
        self.log.debug("Hard flush end")

        if self._perf:
            self.log.debug("Hard flush duration: %s", (time.time() - s))

    #############################################
    # Methods to handle the various insert events
    #############################################
    def workflow(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a workflow insert event.
        """
        wf = self.linedataToObject(linedata, Workflow())
        self.log.trace("Workflow: %s", wf)

        wf.timestamp = wf.ts
        wf.planner_arguments = wf.argv

        is_root = True
        if wf.root_xwf_id != wf.wf_uuid:
            is_root = False
            wf.root_wf_id = self.wf_uuid_to_id(wf.root_xwf_id)

        if wf.parent_wf_id is not None:
            wf.parent_wf_id = self.wf_uuid_to_id(wf.parent_wf_id)

        # workflow inserts must be explicitly written to db whether
        # batching or not
        wf.commit_to_db(self.session)
        if is_root:
            wf.root_wf_id = self.wf_uuid_to_id(wf.root_xwf_id)
            wf.commit_to_db(self.session)
        if wf.root_wf_id is None:
            self.log.warn("Count not determine root_wf_id for event %s", wf)

    def workflowstate(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a workflowstate insert event.
        """
        wfs = self.linedataToObject(linedata, Workflowstate())
        self.log.trace("workflowstate: %s", wfs)

        state = {
            "stampede.xwf.start": "WORKFLOW_STARTED",
            "stampede.xwf.end": "WORKFLOW_TERMINATED",
        }

        wfs.wf_id = self.wf_uuid_to_id(wfs.wf_uuid)
        wfs.timestamp = wfs.ts
        wfs.state = state[wfs.event]

        if self._batch:
            self._batch_cache["batch_events"].append(wfs)
        else:
            wfs.commit_to_db(self.session)
            if wfs.event == "stampede.xwf.end":
                self.purgeCaches(wfs)

    def workflow_meta(self, linedata):
        """

        :param linedata:  dictionary of netlogger BP data dict-ified
        :return:

        Handles a workflow metadata insert event.
        """
        wf_meta = self.linedataToObject(linedata, WorkflowMeta())
        wf_meta.wf_id = self.wf_uuid_to_id(wf_meta.wf_uuid)

        self.log.trace("workflowmeta: %s", wf_meta)

        if self._batch:
            self._batch_cache["batch_events"].append(wf_meta)
        else:
            wf_meta.commit_to_db(self.session)

    def job(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a static job insert event.
        """
        job = self.linedataToObject(linedata, Job())
        job.wf_id = self.wf_uuid_to_id(job.wf_uuid)
        job.clustered = util.as_bool(job.clustered)
        self.log.trace("job: %s", job)

        if self._batch:
            self._batch_cache["batch_events"].append(job)
        else:
            job.commit_to_db(self.session)

    def job_edge(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a static job edge insert event.
        """
        je = self.linedataToObject(linedata, JobEdge())
        je.wf_id = self.wf_uuid_to_id(je.wf_uuid)
        self.log.trace("job_edge: %s", je)

        if self._batch:
            self._batch_cache["batch_events"].append(je)
        else:
            je.commit_to_db(self.session)

    def job_instance(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a job instance insert event.
        """
        job_instance = self.linedataToObject(linedata, JobInstance())
        self.log.trace("job_instance: %s", job_instance)

        job_instance.wf_id = self.wf_uuid_to_id(job_instance.wf_uuid)
        if job_instance.wf_id is None:
            self.log.error(
                "No wf_id associated with wf_uuid %s - can not insert job instance %s",
                job_instance.wf_uuid,
                job_instance,
            )
            return

        job_instance.job_id = self.get_job_id(
            job_instance.wf_id, job_instance.exec_job_id
        )
        if not job_instance.job_id:
            self.log.error(
                "Could not determine job_id for job_instance: %s", job_instance
            )
            return

        if (
            job_instance.event == "stampede.job_inst.submit.start"
            or job_instance.event == "stampede.job_inst.pre.start"
        ):

            iid = self.get_job_instance_id(job_instance, quiet=True)

            if not iid:
                # explicit insert
                job_instance.commit_to_db(self.session)
                # seed the cache
                self.get_job_instance_id(job_instance)

            if job_instance.event == "stampede.job_inst.pre.start":
                self.jobstate(linedata)
            return

        if (
            job_instance.event == "stampede.job_inst.main.end"
            or job_instance.event == "stampede.job_inst.post.end"
        ):

            job_instance.job_instance_id = self.get_job_instance_id(job_instance)

            if self._batch:
                self._batch_cache["update_events"].append(job_instance)
            else:
                job_instance.merge_to_db(self.session)
            self.jobstate(linedata)

    def jobstate(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a jobstate insert event.
        """
        js = self.linedataToObject(linedata, Jobstate())
        self.log.trace("jobstate: %s", js)

        states = {
            # array maps to status [-1, 0]
            "stampede.job_inst.pre.start": [
                "PRE_SCRIPT_STARTED",
                "PRE_SCRIPT_STARTED",
            ],  # statusless
            "stampede.job_inst.pre.term": [
                "PRE_SCRIPT_TERMINATED",
                "PRE_SCRIPT_TERMINATED",
            ],  # s-less
            "stampede.job_inst.pre.end": ["PRE_SCRIPT_FAILED", "PRE_SCRIPT_SUCCESS"],
            "stampede.job_inst.submit.end": ["SUBMIT_FAILED", "SUBMIT"],
            "stampede.job_inst.main.start": ["EXECUTE", "EXECUTE"],  # s-less
            "stampede.job_inst.main.term": ["JOB_EVICTED", "JOB_TERMINATED"],
            "stampede.job_inst.main.end": ["JOB_FAILURE", "JOB_SUCCESS"],
            "stampede.job_inst.post.start": [
                "POST_SCRIPT_STARTED",
                "POST_SCRIPT_STARTED",
            ],  # s-less
            "stampede.job_inst.post.term": [
                "POST_SCRIPT_TERMINATED",
                "POST_SCRIPT_TERMINATED",
            ],  # s-less
            "stampede.job_inst.post.end": ["POST_SCRIPT_FAILED", "POST_SCRIPT_SUCCESS"],
            "stampede.job_inst.held.start": ["JOB_HELD", "JOB_HELD"],  # s-less
            "stampede.job_inst.held.end": ["JOB_RELEASED", "JOB_RELEASED"],  # s-less
            "stampede.job_inst.image.info": ["IMAGE_SIZE", "IMAGE_SIZE"],  # s-less
            "stampede.job_inst.abort.info": ["JOB_ABORTED", "JOB_ABORTED"],  # s-less
            "stampede.job_inst.grid.submit.end": ["GRID_SUBMIT_FAILED", "GRID_SUBMIT"],
            "stampede.job_inst.globus.submit.end": [
                "GLOBUS_SUBMIT_FAILED",
                "GLOBUS_SUBMIT",
            ],
        }

        if js.event not in states:
            # corner case event
            js.state = js.event.split(".")[2].upper()
        else:
            # doctor status-less events to simplify code
            if not hasattr(js, "status"):
                js.status = 0
            js.state = states[js.event][int(js.status) + 1]

        js.job_instance_id = self.get_job_instance_id(js)
        if not js.job_instance_id:
            self.log.error("No job_instance_id for event: %s -%s", linedata, js)
            return

        js.timestamp = js.ts

        if self._batch:
            self._batch_cache["batch_events"].append(js)
        else:
            js.commit_to_db(self.session)

    def invocation(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a invocation insert event.
        """
        invocation = self.linedataToObject(linedata, Invocation())
        self.log.trace("invocation: %s", invocation)

        invocation.wf_id = self.wf_uuid_to_id(invocation.wf_uuid)

        invocation.job_instance_id = self.get_job_instance_id(invocation)
        if invocation.job_instance_id is None:
            self.log.error(
                "Could not determine job_instance_id for invocation: %s", invocation
            )
            return

        if self._batch:
            self._batch_cache["batch_events"].append(invocation)
        else:
            invocation.commit_to_db(self.session)

    def task(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a task insert event
        """
        task = self.linedataToObject(linedata, Task())
        self.log.trace("task: %s", task)
        task.wf_id = self.wf_uuid_to_id(task.wf_uuid)

        if self._batch:
            self._batch_cache["batch_events"].append(task)
        else:
            task.commit_to_db(self.session)

    def task_edge(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a task edge insert event
        """
        if linedata["xwf.id"] not in self._task_edge_flush:
            if self._batch:
                self.hard_flush()
            self._task_edge_flush[linedata["xwf.id"]] = True

        te = self.linedataToObject(linedata, TaskEdge())
        self.log.trace("task_event: %s", te)
        te.wf_id = self.wf_uuid_to_id(te.wf_uuid)

        if self._batch:
            self._batch_cache["batch_events"].append(te)
        else:
            te.commit_to_db(self.session)

    def task_map(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of DB data dict-ified

        Handles a task.map event.  Updates a Task table row
        to include the proper job_id event.
        """
        # Flush previous events to ensure that all the batched
        # Job table entries are written.
        if linedata["xwf.id"] not in self._task_map_flush:
            if self._batch:
                self.hard_flush()
            self._task_map_flush[linedata["xwf.id"]] = True

        wf_id = self.wf_uuid_to_id(linedata["xwf.id"])
        job_id = self.get_job_id(wf_id, linedata["job.id"])

        if not job_id:
            self.log.error("Could not determine job_id for task map: %s", linedata)
            return

        try:
            task = (
                self.session.query(Task)
                .filter(Task.wf_id == wf_id)
                .filter(Task.abs_task_id == linedata["task.id"])
                .one()
            )
            task.job_id = job_id
        except orm.exc.MultipleResultsFound as e:
            self.log.error("Multiple task results: cant map task: %s ", linedata)
            return
        except orm.exc.NoResultFound as e:
            self.log.error("No task found: cant map task: %s ", linedata)
            return

        if self._batch:
            # next flush will catch this - no cache
            pass
        else:
            self.session.commit()

    def task_meta(self, linedata):
        """

        :param linedata:  dictionary of netlogger BP data dict-ified
        :return:

        Handles a task metadata insert event.
        """
        task_meta = self.linedataToObject(linedata, TaskMeta())
        task_meta.wf_id = self.wf_uuid_to_id(task_meta.wf_uuid)
        task_meta.task_id = self.get_task_id(task_meta.wf_id, task_meta.abs_task_id)

        self.log.trace("task_meta: %s", task_meta)

        if self._batch:
            self._batch_cache["batch_events"].append(task_meta)
        else:
            task_meta.commit_to_db(self.session)

    def rc_meta(self, linedata):
        """

        :param linedata:  dictionary of netlogger BP data dict-ified
        :return:

        Handles a rc metadata insert event.
        """
        rc_meta = self.linedataToObject(linedata, RCMeta())
        lfn = rc_meta.lfn_id

        # str’ object has no attribute ‘_sa_instance_state’ error if the below is set
        # in 5.0 schema lfn maps to RCLFN object based on how the sql alchemy schema is defined
        # rc_meta.lfn = lfn

        rc_meta.wf_id = self.wf_uuid_to_id(rc_meta.wf_uuid)
        rc_meta.lfn_id = self.get_lfn_id(rc_meta.wf_id, lfn)

        self.log.trace("rc_meta: %s", rc_meta)

        # we have to do the merge individually to prevent integrity constraint
        # errors that happen if we put them in the batch cache update_events
        rc_meta.merge_to_db(self.session)

    def int_metric(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a integrity metric event
        """
        int_meta = self.linedataToObject(linedata, IntegrityMetrics())
        self.log.trace("int_meta: %s", int_meta)

        int_meta.wf_id = self.wf_uuid_to_id(int_meta.wf_uuid)

        int_meta.job_instance_id = self.get_job_instance_id(int_meta)
        if int_meta.job_instance_id is None:
            self.log.error(
                "Could not determine job_instance_id for int_meta: %s", int_meta
            )
            return

        if self._batch:
            self._batch_cache["batch_events"].append(int_meta)
        else:
            int_meta.commit_to_db(self.session)

    def tag(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a job_instance tag event
        """
        tag = self.linedataToObject(linedata, Tag())
        self.log.trace("job_inst.tag: %s", tag)

        tag.wf_id = self.wf_uuid_to_id(tag.wf_uuid)

        tag.job_instance_id = self.get_job_instance_id(tag)
        if tag.job_instance_id is None:
            self.log.error("Could not determine job_instance_id for tag: %s", tag)
            return

        if self._batch:
            self._batch_cache["batch_events"].append(tag)
        else:
            tag.commit_to_db(self.session)

    def rc_pfn(self, linedata):
        """

        :param linedata:  dictionary of netlogger BP data dict-ified
        :return:

        Handles a rc pfn insert event that populates the pfn and the site attribute.
        """
        rc_pfn = self.linedataToObject(linedata, RCPFN())
        lfn = rc_pfn.lfn_id
        # str’ object has no attribute ‘_sa_instance_state’ error if the below is set
        # in 5.0 schema lfn maps to RCLFN object based on how the sql alchemy schema is defined
        # rc_pfn.lfn = lfn
        rc_pfn.wf_id = self.wf_uuid_to_id(rc_pfn.wf_uuid)
        rc_pfn.lfn_id = self.get_lfn_id(rc_pfn.wf_id, lfn)

        self.log.trace("rc_pfn: %s", rc_pfn)

        if self._batch:
            self._batch_cache["batch_events"].append(rc_pfn)
        else:
            rc_pfn.commit_to_db(self.session)

    def wf_task_file_map(self, linedata):
        """
        Handles the event that associates workflow, task and a file lfn .
        Populates to wf_files

        :param linedata:
        :return:
        """

        wf_files = self.linedataToObject(linedata, WorkflowFiles())
        lfn = wf_files.lfn_id
        # str’ object has no attribute ‘_sa_instance_state’ error if the below is set
        # in 5.0 schema lfn maps to RCLFN object based on how the sql alchemy schema is defined
        # wf_files.lfn = lfn
        wf_files.wf_id = self.wf_uuid_to_id(wf_files.wf_uuid)
        wf_files.task_id = self.get_task_id(wf_files.wf_id, wf_files.abs_task_id)
        wf_files.lfn_id = self.get_lfn_id(wf_files.wf_id, lfn)

        self.log.trace("wf_files: %s", wf_files)

        if self._batch:
            self._batch_cache["batch_events"].append(wf_files)
        else:
            wf_files.commit_to_db(self.session)

    def subwf_map(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a subworkflow job map event.
        """
        self.log.trace("subwf_map: %s", linedata)

        wf_id = self.wf_uuid_to_id(linedata["xwf.id"])
        subwf_id = self.wf_uuid_to_id(linedata["subwf.id"])
        job_id = self.get_job_id(wf_id, linedata["job.id"])

        try:
            job_inst = (
                self.session.query(JobInstance)
                .filter(JobInstance.job_id == job_id)
                .filter(JobInstance.job_submit_seq == linedata["job_inst.id"])
                .one()
            )
            job_inst.subwf_id = subwf_id
        except orm.exc.MultipleResultsFound as e:
            self.log.error(
                "Multiple job instance results: cant map subwf: %s ", linedata
            )
            return
        except orm.exc.NoResultFound as e:
            self.log.error("No job instance found: cant map subwf: %s ", linedata)
            return

        if self._batch:
            # next flush will catch this - no cache
            pass
        else:
            self.session.commit()

    def host(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a host insert event.
        """
        host = self.linedataToObject(linedata, Host())

        self.log.trace("host: %s", host)

        if self.hosts_written_cache is None:
            self.hosts_written_cache = {}
            query = self.session.query(Host)
            for row in query.all():
                self.hosts_written_cache[
                    (row.wf_id, row.site, row.hostname, row.ip)
                ] = True

        host.wf_id = self.wf_uuid_to_root_id(host.wf_uuid)

        # handle inserts into the host table
        if (
            host.wf_id,
            host.site,
            host.hostname,
            host.ip,
        ) not in self.hosts_written_cache:
            if self._batch:
                self._batch_cache["batch_events"].append(host)
            else:
                host.commit_to_db(self.session)
            self.hosts_written_cache[
                (host.wf_id, host.site, host.hostname, host.ip)
            ] = True

        # handle mappings
        if self._batch:
            self._batch_cache["host_map_events"].append(host)
        else:
            self.map_host_to_job_instance(host)

    def static_end(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        This forces a flush after all of the static events
        have been processed.
        """
        self.log.trace("static_end: %s", linedata)
        if self._batch:
            self.hard_flush()

    def static_meta_start(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        This forces a flush of all the abstract related events,
        so that the task id's can be retrieved for metadata population
        of tasks.
        """
        self.log.trace("static_meta_start: %s", linedata)
        if self._batch:
            self.hard_flush()

    def noop(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        A NOOP method for events that are being ignored.
        """
        self.log.trace("noop: %s", linedata)

    ####################################
    # DB helper/lookup/caching functions
    ####################################
    def wf_uuid_to_id(self, wf_uuid):
        """
        @type   wf_uuid: string
        @param  wf_uuid: wf_uuid string from BP logs

        Attempts to retrieve a workflow wf_id PK/FK from cache.  If
        not in cache, retrieve from st_workflow table in DB and cache.
        Cuts down on DB queries during insert processing.
        """
        if wf_uuid not in self.wf_id_cache:
            query = self.session.query(Workflow).filter(Workflow.wf_uuid == wf_uuid)
            try:
                self.wf_id_cache[wf_uuid] = query.one().wf_id
            except orm.exc.MultipleResultsFound as e:
                self.log.error("Multiple wf_id results for wf_uuid %s : %s", wf_uuid, e)
                return None
            except orm.exc.NoResultFound as e:
                self.log.error("No wf_id results for wf_uuid %s : %s", wf_uuid, e)
                return None

        return self.wf_id_cache[wf_uuid]

    def wf_uuid_to_root_id(self, wf_uuid):
        """
        @type   wf_uuid: string
        @param  wf_uuid: wf_uuid string from BP logs

        Attempts to retrieve a root workflow wf_id PK/FK from cache.  If
        not in cache, retrieve from st_workflow table in DB and cache.
        Cuts down on DB queries during insert processing.
        """
        if wf_uuid not in self.root_wf_id_cache:
            query = self.session.query(Workflow).filter(Workflow.wf_uuid == wf_uuid)
            try:
                self.root_wf_id_cache[wf_uuid] = query.one().root_wf_id
            except orm.exc.MultipleResultsFound as e:
                self.log.error("Multiple wf_id results for wf_uuid %s : %s", wf_uuid, e)
                return None
            except orm.exc.NoResultFound as e:
                self.log.error("No wf_id results for wf_uuid %s : %s", wf_uuid, e)
                return None

        return self.root_wf_id_cache[wf_uuid]

    def get_task_id(self, wf_id, task_dax_id):
        """
        @type   wf_id: int
        @param  wf_id: A workflow id from the workflow table.
        @type   task_dax_id: string
        @param  task_dax_id: The ID for the task in the DAX

        Gets and caches task_id for task_meta inserts
        """
        if (wf_id, task_dax_id) not in self.task_id_cache:
            query = (
                self.session.query(Task.task_id)
                .filter(Task.wf_id == wf_id)
                .filter(Task.abs_task_id == task_dax_id)
            )
            try:
                self.task_id_cache[((wf_id, task_dax_id))] = query.one().task_id
            except orm.exc.MultipleResultsFound as e:
                self.log.error(
                    "Multiple results found for wf_uuid/task_dax_id: %s/%s",
                    wf_id,
                    task_dax_id,
                )
                return None
            except orm.exc.NoResultFound as e:
                self.log.error(
                    "No results found for wf_uuid/task_dax_id: %s/%s",
                    wf_id,
                    task_dax_id,
                )
                return None

        return self.task_id_cache[(wf_id, task_dax_id)]

    def get_lfn_id(self, wf_id, lfn):
        """
        @type   wf_id: int
        @param  wf_id: A workflow id from the workflow table.
        @type   lfn: string
        @param  lfn: The logical filename for the file

        Gets and caches lfn_id for rc_meta, rc_lfn, rc_pfn and wf_files inserts
        """
        if (wf_id, lfn) not in self.lfn_id_cache:
            id = self.__get_lfn_id_from_database__(wf_id, lfn)

            if id is None:
                # we do an explicit insert to populate the RCLFN table
                file = RCLFN()
                file.lfn = lfn
                # explicit insert
                file.commit_to_db(self.session)
                # retrieve from the database and set the cache
                id = self.__get_lfn_id_from_database__(wf_id, lfn)

                # if ID is still None then definitely an an error
                if id is None:
                    self.log.error(
                        "No results found for wf_uuid/lfn: %s/%s", wf_id, lfn
                    )
                    return None

            self.lfn_id_cache[(wf_id, lfn)] = id

        return self.lfn_id_cache[(wf_id, lfn)]

    def __get_lfn_id_from_database__(self, wf_id, lfn):
        """
        @type   wf_id: int
        @param  wf_id: A workflow id from the workflow table.
        @type   lfn: string
        @param  lfn: The logical filename for the file

        Retrieves the LFN explicitly by querying the database.
        """

        query = self.session.query(RCLFN.lfn_id).filter(RCLFN.lfn == lfn)
        lfn_id = None
        try:
            lfn_id = query.one().lfn_id
        except orm.exc.MultipleResultsFound as e:
            self.log.error("Multiple results found for wf_uuid/lfn: %s/%s", wf_id, lfn)
            return None
        except orm.exc.NoResultFound as e:
            return None

        return lfn_id

    def get_job_id(self, wf_id, exec_id):
        """
        @type   wf_id: int
        @param  wf_id: A workflow id from the workflow table.
        @type   exec_id: string
        @param  exec_id: The exec_job_id for a given job.

        Gets and caches job_id for job_instance inserts and static
        table updating.
        """
        if (wf_id, exec_id) not in self.job_id_cache:
            query = (
                self.session.query(Job.job_id)
                .filter(Job.wf_id == wf_id)
                .filter(Job.exec_job_id == exec_id)
            )
            try:
                self.job_id_cache[((wf_id, exec_id))] = query.one().job_id
            except orm.exc.MultipleResultsFound as e:
                self.log.error(
                    "Multiple results found for wf_uuid/exec_job_id: %s/%s",
                    wf_id,
                    exec_id,
                )
                return None
            except orm.exc.NoResultFound as e:
                self.log.error(
                    "No results found for wf_uuid/exec_job_id: %s/%s", wf_id, exec_id
                )
                self.log.error(query)
                return None

        return self.job_id_cache[((wf_id, exec_id))]

    def get_job_instance_id(self, o, quiet=False):
        """
        @type   o: class instance
        @param  o: Mapper object containing wf_uuid and exec_job_id.

        Attempts to retrieve a job job_instance_id PK/FK from cache.  If not in
        cache, retrieve from st_job table.
        """
        wf_id = self.wf_uuid_to_id(o.wf_uuid)
        cached_job_id = self.get_job_id(wf_id, o.exec_job_id)
        uniqueIdIdx = (cached_job_id, o.job_submit_seq)
        if uniqueIdIdx not in self.job_instance_id_cache:
            query = (
                self.session.query(JobInstance)
                .filter(JobInstance.job_id == cached_job_id)
                .filter(JobInstance.job_submit_seq == o.job_submit_seq)
            )
            try:
                self.job_instance_id_cache[uniqueIdIdx] = query.one().job_instance_id
            except orm.exc.MultipleResultsFound as e:
                if not quiet:
                    self.log.error(
                        "Multple job_instance_id results for tuple %s : %s",
                        uniqueIdIdx,
                        e,
                    )
                return None
            except orm.exc.NoResultFound as e:
                if not quiet:
                    self.log.error(
                        "No job_instance_id results for tuple %s : %s", uniqueIdIdx, e
                    )
                return None

        return self.job_instance_id_cache[uniqueIdIdx]

    def map_host_to_job_instance(self, host):
        """
        @type   host: class instance of stampede_schema.Host
        @param  host: Host object with info from a host event in the log

        A single job may have multiple (redundant) host events.  This
        checks the cache to see if a job had already had its host_id,
        and if not, do the proper update and note it in the cache.
        """
        self.log.trace("map_host_to_job_instance: %s", host)

        wf_id = self.wf_uuid_to_id(host.wf_uuid)
        cached_job_id = self.get_job_id(wf_id, host.exec_job_id)

        if (cached_job_id, host.job_submit_seq) not in self.host_cache:
            if not host.host_id:
                try:
                    host.host_id = (
                        self.session.query(Host.host_id)
                        .filter(Host.wf_id == host.wf_id)
                        .filter(Host.site == host.site)
                        .filter(Host.hostname == host.hostname)
                        .filter(Host.ip == host.ip)
                        .one()
                        .host_id
                    )
                except orm.exc.MultipleResultsFound as e:
                    self.log.error("Multiple host_id results for host: %s", host)
            job_instance = (
                self.session.query(JobInstance)
                .filter(JobInstance.job_id == cached_job_id)
                .filter(JobInstance.job_submit_seq == host.job_submit_seq)
                .one()
            )
            job_instance.host_id = host.host_id
            job_instance.merge_to_db(self.session, batch=self._batch)
            self.host_cache[(cached_job_id, host.job_submit_seq)] = True

    def purgeCaches(self, wfs):
        """
        @type   wfs: class instance of stampede_schema.Workflowstate
        @param  wfs: Workflow state object from an end event.

        Purges information from the lookup caches after a workflow.end
        event has been recieved.
        """
        self.log.debug("Purging caches for: %s", wfs.wf_uuid)

        self.purgeCache(self.wf_id_cache, wfs.wf_uuid)

        self.purgeCache(self.root_wf_id_cache, wfs.wf_uuid)

        self.purgeCache(self.job_instance_id_cache, wfs.wf_id)

        self.purgeCache(self.host_cache, wfs.wf_uuid)

        self.purgeCache(self.task_id_cache, wfs.wf_id)

        self.purgeCache(self.lfn_id_cache, wfs.wf_id)

        self.purgeCache(self.job_id_cache, wfs.wf_id)

        if wfs.wf_uuid in self._task_map_flush:
            del self._task_map_flush[wfs.wf_uuid]

    def purgeCache(self, cache, key):
        """
        Removes from a cache an entry matching a key id
        :param cache:
        :param key_id:
        :return:
        """
        if key in cache:
            del cache[key]

    ################
    # Cleanup, etc
    ################

    def finish(self):
        if self._batch:
            self.log.info("Executing final flush")
            self.hard_flush()
        self.disconnect()
        if self._perf:
            run_time = time.time() - self._start_time
            self.log.info(
                "Loader performance: insert_time=%s, insert_num=%s, "
                "total_time=%s, run_time_delta=%s, mean_time=%s",
                self._insert_time,
                self._insert_num,
                run_time,
                run_time - self._insert_time,
                self._insert_time / self._insert_num,
            )



"""
NetLogger interactions with the Python logging module.
"""
__rcsid__ = "$Id: logutil.py 772 2008-05-23 22:59:22Z dang $"

import logging
import logging.handlers
import optparse
import os
import sys
import time
import traceback
import types

from Pegasus.netlogger import nlapi
from Pegasus.netlogger.nlapi import Level
from Pegasus.netlogger.version import *

# extra logging levels
TRACE = logging.DEBUG - 1

# Top-level qualified name for netlogger
PROJECT_NAMESPACE = "netlogger"

# Global holder of the "correct" NetLogger class
# to use when instantiating new loggers
_logger_class = None


def setLoggerClass(clazz):
    """Set the class used by NetLogger logging"""
    global _logger_class
    _logger_class = clazz


# consistent with new naming style
set_logger_class = setLoggerClass


def get_logger(filename):
    """
    Return a NetLogger logger with qualified name based on the provied
    filename.  This method is indended to be called by scripts and
    modules by passing in their own __file__ as filename after already
    having initialized the logging module via the NL OptionParser or
    some equivalent action.

    If the logger name starts with a '.', it will be taken as-is, with
    the leading '.' stripped.
    Otherwise, the logger will be rooted at PROJECT_NAMESPACE.

    Parameters:
    filename - The full filename of the NL script or module requesting
    a logger, i.e. __file__
    """
    if filename == "":
        qname = ""
    elif filename[0] == ".":
        qname = filename
    else:
        qname = ".".join(_modlist(filename))
    return _logger(qname)


def get_root_logger():
    """Return root for all NetLogger loggers."""
    return _logger("")  # logging.getLogger(PROJECT_NAMESPACE)


def _logger(qualname):
    """
    Return a logger based on the provided qualified name
    Prepend PROJECT_NAMESPACE if not already there, unless
    qualified name starts with a '.'.
    """
    # Mess with qualified name
    if not qualname:
        qualname = PROJECT_NAMESPACE
    elif qualname[0] == ".":
        qualname = qualname[1:]
    elif not qualname.startswith(PROJECT_NAMESPACE):
        qualname = PROJECT_NAMESPACE + "." + qualname
    # Swap in "my" logger class, create logger, swap back out
    orig_class = logging.getLoggerClass()
    logging.setLoggerClass(_logger_class)
    logger = logging.getLogger(qualname)
    logging.setLoggerClass(orig_class)
    # Return "my" new logger instance
    return logger


def _modlist(path):
    """
    Return a list of module names based on the path provided.  The
    expected path list will be rooted at either "netlogger" or
    "scripts" so won't contain either as one of the module names.  Any
    tailing python extension is also trimmed.
    """
    if path == "/":
        return []
    head, tail = os.path.split(path)
    # ignore python extensions
    if tail.endswith(".py") or tail.endswith(".pyc"):
        tail = os.path.splitext(tail)[0]
    # stop if at top of source tree
    if tail in ("netlogger", "scripts"):
        return []
    # stop if at root of path
    if head == "" or head == ".":
        return [tail]
    # otherwise continue
    return _modlist(head) + [tail]


class DoesLogging:
    """Mix-in class that creates the attribute 'log', setting its qualified
    name to the name of the module and class.
    """

    def __init__(self, name=None):
        if name is None:
            if self.__module__ != "__main__":
                name = "{}.{}".format(self.__module__, self.__class__.__name__)
            else:
                name = self.__class__.__name__
        self.log = _logger(name)
        # cache whether log is debug or higher in a flag to
        # lower overhead of debugging statements
        self._dbg = self.log.isEnabledFor(logging.DEBUG)
        self._trace = self.log.isEnabledFor(TRACE)


class BPLogger(logging.Logger):
    """Logger class that writes Best-Practices formatted logs.

    Usage:
        The arguments are not exactly the same as for the Logger in
        the logging module. Instead they consist of an event name
        and keywords that are name=value pairs for the event contents.

        An exception to this is the exc() or exception() method,
        which takes an Exception instance as its second argument
        in addition to the event name.

    Example:
        log.info("my.event", value=88.7, units="seconds")
        # output
        # ts=2009-07-24T20:18:04.775650Z event=netlogger.my.event level=INFO units=seconds value=88.700000

    """

    def __init__(self, qualname):
        self._qualname = qualname
        self._format = nlapi.Log(newline=False, level=nlapi.Level.ALL)
        logging.Logger.__init__(self, qualname)

    def set_meta(self, **kw):
        """Set metadata to be logged with every event, e.g.
        an identifier or host name.
        """
        self._format.setMeta(None, **kw)

    def log(self, level, nl_level, event, exc_info=None, **kwargs):
        ts = time.time()
        if self._qualname:
            event = self._qualname + "." + event
        # replace '__' with '.'
        kwargs = {key.replace("__", "."): value for key, value in kwargs.iteritems()}
        # format as BP
        msg = self._format(event, ts, nl_level, **kwargs)
        logging.Logger.log(self, level, msg, exc_info=exc_info)

    def trace(self, event, **kwargs):
        if self.isEnabledFor(TRACE):
            self.log(TRACE, Level.TRACE, event, **kwargs)

    def debug(self, event, **kwargs):
        if self.isEnabledFor(logging.DEBUG):
            self.log(logging.DEBUG, Level.DEBUG, event, **kwargs)

    def info(self, event, **kwargs):
        self.log(logging.INFO, Level.INFO, event, **kwargs)

    def warning(self, event, **kwargs):
        self.log(logging.WARN, Level.WARN, event, **kwargs)

    warn = warning

    def error(self, event, **kwargs):
        self.log(logging.ERROR, Level.ERROR, event, **kwargs)

    def critical(self, event, **kwargs):
        self.log(logging.CRITICAL, Level.FATAL, event, **kwargs)

    def exception(self, event, err, **kwargs):
        estr = traceback.format_exc()
        estr = " | ".join(e.strip() for e in estr.split("\n"))
        self.log(
            logging.ERROR,
            Level.ERROR,
            event,
            msg=str(err),
            status=-1,
            traceback=estr,
            **kwargs,
        )

    exc = exception


class BPSysLogger(BPLogger):
    """This is a hack that prepends a header to the
    output of BPLogger in order to work-around some bug with the
    Python SysLogHandler and Ubuntu rsylog that otherwise splits
    out the first section of the timestamp as part of the header.
    """

    header = "netlogger"  # header prefix

    def __init__(self, qualname):
        BPLogger.__init__(self, qualname)
        self._orig_format = self._format
        self._format = self.syslog_format
        self._hdr = self.header + ": "

    def set_meta(self, **kw):
        """See set_meta() in superclass.

        Repeated here because superclass method accesses a protected
        attribute that was modified in the constructor.
        """
        self._orig_format.setMeta(None, **kw)

    def flush(self):
        self._orig_format.flush()

    def syslog_format(self, *arg, **kw):
        return self._hdr + self._orig_format(*arg, **kw)


###############################################
## Set BPLogger as default logging class
###############################################
setLoggerClass(BPLogger)


class PrettyBPLogger(BPLogger):
    """Logger class that writes 'pretty' Best-Practices formatted logs.
    This is a variation on BP format. Stack traces logged with the
    method exc() or exception() will be in their original form.

    Usage:
        See Usage notes for BPLogger.

    Example:
        log.info("my.event", value=88.7, units="seconds")
        # output
        # 2009-07-24T20:18:04.716913Z INFO  netlogger.my.event - units=seconds,value=88.7

    """

    def __init__(self, qualname):
        BPLogger.__init__(self, qualname)
        self._format = nlapi.Log(newline=False, level=nlapi.Level.ALL, pretty=True)

    def exception(self, event, err, **kwargs):
        tbstr = traceback.format_exc()
        self.log(logging.ERROR, Level.ERROR, event, traceback=tbstr, **kwargs)

    exc = exception


class RawBPLogger(logging.Logger):
    """Logger class that does not modify the message, just leaves
    it as a 'raw' dictionary. This is useful for network communication
    that is just going to pickle the event anyways.
    """

    def log(self, level, nl_level, event, exc_info=None, **kwargs):
        ts = time.time()
        if self._qualname:
            event = self._qualname + "." + event
        # replace '__' with '.'
        kwargs = {key.replace("__", "."): value for key, value in kwargs.iteritems()}
        # build msg dictionary
        msg = {"event": event, "ts": ts, "level": nl_level}
        msg.update(kwargs)
        # 'write' out
        logging.Logger.log(self, level, msg, exc_info=exc_info)


class FakeBPLogger(logging.Logger):
    def __init__(self, qualname):
        logging.Logger.__init__(self, qualname)

    def set_meta(self, **kw):
        pass

    def log(self, level, nl_level, event, **kwargs):
        pass

    def trace(self, event, **kwargs):
        pass

    def debug(self, event, **kwargs):
        pass

    def info(self, event, **kwargs):
        pass

    def warning(self, event, **kwargs):
        pass

    warn = warning

    def error(self, event, **kwargs):
        pass

    def critical(self, event, **kwargs):
        pass

    def exception(self, event, err, **kwargs):
        pass

    exc = exception


def profile(func):
    """decorator for start,end event function profiling with netlogger."""
    if os.getenv("NETLOGGER_ON", False) in ("off", "0", "no", "false", "", False):
        return func
    if not isinstance(func, types.FunctionType):
        return func

    if func.__module__ == "__main__":
        f = func.__globals__["__file__"] or "unknown"
        event = "%s" % os.path.splitext(os.path.basename(f))[0]
        log = _logger("script")
        log.set_meta(file=f, pid=os.getpid(), ppid=os.getppid(), gpid=os.getgid())
    else:
        event = "%s" % func.__name__
        log = _logger("%s" % func.__module__)
        log.set_meta(pid=os.getpid(), ppid=os.getppid(), gpid=os.getgid())

    def nl_profile_func(*args, **kw):
        log.debug("%s.start" % event)
        try:
            v = func(*args, **kw)
        except Exception:
            log.error("%s.end" % event)
            raise
        log.debug("%s.end" % event)
        return v

    return nl_profile_func


def profile_result(func):
    """decorator for start,end event function profiling with netlogger.
    return value is logged as result.
    """
    if os.getenv("NETLOGGER_ON", False) in ("off", "0", "no", "false", "", False):
        return func
    if not isinstance(func, types.FunctionType):
        return func

    if func.__module__ == "__main__":
        f = func.__globals__["__file__"] or "unknown"
        event = "%s" % os.path.splitext(os.path.basename(f))[0]
        log = _logger("script")
        log.set_meta(file=f, pid=os.getpid(), ppid=os.getppid(), gpid=os.getgid())
    else:
        event = "%s" % func.__name__
        log = _logger("%s" % func.__module__)
        log.set_meta(pid=os.getpid(), ppid=os.getppid(), gpid=os.getgid())

    def nl_profile_func(*args, **kw):
        log.debug("%s.start" % event)
        try:
            v = func(*args, **kw)
        except Exception:
            log.error("%s.end" % event)
            raise
        log.debug("%s.end" % event, result=v)
        return v

    return nl_profile_func


class Profiler(type):
    """metaclass that will wrap all user defined methods with start and end event logs.
    Currently wrapping only instancemethod type.

    Variables:
        profiler_skip_methods: list of methods profiler will skip
        profile_skip_all: profiler will not wrap any methods
        _log: Logging object to use with class

    Usage:
        class MyClass:
            __metaclass__ = Profiler
            profiler_skip_methods = ['__init__', 'getsecret']
            profiler_skip_all = False
    """

    profiler_skip_methods = ["__init__"]
    profiler_skip_all = False

    @staticmethod
    def __profile_method(func):
        """decorator for start,end event method profiling with netlogger
        skips any classmethod or staticmethod types.
        """
        if not isinstance(func, types.FunctionType):
            return func

        event = "%s" % func.__name__

        def nl_profile_method(self, *args, **kw):
            self._log.debug("%s.start" % event)
            try:
                v = func(self, *args, **kw)
            except Exception:
                self._log.error("%s.end" % event)
                raise
            self._log.debug("%s.end" % event)
            return v

        return nl_profile_method

    def __new__(cls, classname, bases, classdict):
        if os.getenv("NETLOGGER_ON", False) in ("off", "0", "no", "false", "", False):
            setLoggerClass(FakeBPLogger)
            classdict["_log"] = _logger(
                "{}.{}".format(classdict["__module__"], classname)
            )
            return type.__new__(cls, classname, bases, classdict)

        classdict["_log"] = log = _logger(
            "{}.{}".format(classdict["__module__"], classname)
        )
        log.set_meta(pid=os.getpid(), ppid=os.getppid(), gpid=os.getgid())
        keys = []
        if not classdict.get("profiler_skip_all", cls.profiler_skip_all):
            keys = [
                k
                for k in classdict.keys()
                if isinstance(classdict[k], types.FunctionType)
                and k
                not in classdict.get("profiler_skip_methods", cls.profiler_skip_methods)
            ]

        for k in keys:
            classdict[k] = cls.__profile_method(classdict[k])

        return type.__new__(cls, classname, bases, classdict)


class MethodProfiler(Profiler):
    """metaclass that will wrap all user defined methods with start and end event logs.
    Currently wrapping only instancemethod type.
    """

    profiler_skip_all = False


class BasicProfiler(Profiler):
    """metaclass does not wrap methods with 'start' and 'end' tags, to do that use 'Profiler'.
    Useful for classes where one only wants to do 'precision' logging.
    """

    profiler_skip_all = True


class OptionParser(optparse.OptionParser):
    """Set logging 'tude for scripts.

    Usage:
       parser = NLOptionParser(..., can_be_daemon=True/False)
       # add rest of options to 'parser'...
       # This next line sets up logging as a side-effect
       parser.parse_args()
       # rest of program ..

    *******************************************************
    | Pseudo-code description of logic to determine which |
    | types of logs to produce, and where to send them    |
    *******************************************************
    Variables:
        D - daemon mode [True | False]
        L - log file [Missing | Empty | filename]
    Logic:
      if (D) then
          case (L = Missing)
            error!
          case (L = Empty)
            error!
          case (L = filename)
            stderr -> filename
            BP logs -> filename
      else
          case (L = Missing)
            stderr -> stderr
            Pretty logs -> stderr
          case (L = Empty)
            stderr -> stderr
            BP logs -> stderr
          case (L = filename)
            stderr -> stderr
            BP logs -> filename

    *******************************************************
    | Pseudo-code description of logic for verbosity      |
    *******************************************************
    Variables:
        V - verbosity [0 .. N]
        Q - quiet [True | False]
    Logic:
        if (Q) then
            case (V > 0)
                error!
            else
                set verbosity -> OFF
        else
            case V = 0
                set verbosity -> WARN
            case V = 1
                set verbosity -> INFO
            case V = 2
                set verbosity -> DEBUG
            case V >= 3
                set verbosity -> TRACE

    """

    # Attribute (option parser 'dest') names
    DEST_LOG = "log_file"
    DEST_VERBOSE = "verbose"
    DEST_QUIET = "quiet"
    DEST_DAEMON = "daemon"
    DEST_ROT = "log_rotate"

    # Option names, by attribute
    OPTIONS = {
        DEST_LOG: ("-L", "--log"),
        DEST_VERBOSE: ("-v", "--verbose"),
        DEST_QUIET: ("-q", "--quiet"),
        DEST_DAEMON: (None, "--daemon"),
        DEST_ROT: ("-R", "--logrotate"),
    }

    # Verbosity (number of -v's) to logging level
    VBMAP = (logging.WARN, logging.INFO, logging.DEBUG, TRACE)

    def __init__(self, can_be_daemon=False, **kwargs):
        """Add logging-related command-line options
        to an option parser.

        Parameters:
            can_be_daemon - if True, add an option for daemonizing
            kwargs - additional keywords for OptionParser.
                     The 'version' argument will override the default
                     version
        """
        if "version" not in kwargs:
            version_str = "%prog, NetLogger Toolkit version: {}\n  {}".format(
                NL_VERSION, NL_CREATE_DATE,
            )
            version_str += "\n\n" + NL_COPYRIGHT
            kwargs["version"] = version_str
        optparse.OptionParser.__init__(self, **kwargs)
        self._dmn = can_be_daemon

    def _add_options(self):
        group = optparse.OptionGroup(self, "Logging options")
        if self._dmn:
            self.add_option(
                self.OPTIONS[self.DEST_DAEMON][1],
                action="store_true",
                dest=self.DEST_DAEMON,
                default=False,
                help="run in daemon mode",
            )
            logfile_default = "required"
        else:
            logfile_default = "default=stderr"
        group.add_option(
            self.OPTIONS[self.DEST_LOG][0],
            self.OPTIONS[self.DEST_LOG][1],
            default=None,
            action="store",
            dest=self.DEST_LOG,
            metavar="FILE",
            help="write logs to FILE (%s)" % logfile_default,
        )
        group.add_option(
            self.OPTIONS[self.DEST_ROT][0],
            self.OPTIONS[self.DEST_ROT][1],
            default=None,
            action="store",
            dest=self.DEST_ROT,
            metavar="TIME",
            help="rotate logs at an interval (<N>d or <N>h or <N>m)",
        )
        group.add_option(
            self.OPTIONS[self.DEST_VERBOSE][0],
            self.OPTIONS[self.DEST_VERBOSE][1],
            action="count",
            default=0,
            dest=self.DEST_VERBOSE,
            help="more verbose logging",
        )
        group.add_option(
            self.OPTIONS[self.DEST_QUIET][0],
            self.OPTIONS[self.DEST_QUIET][1],
            action="store_true",
            default=False,
            dest=self.DEST_QUIET,
            help="quiet mode, no logging",
        )
        self.add_option_group(group)

    def check_required(self, opt):
        """Simplify checks for required values.
        The default value for a required option must be None.
        The easiest way to achieve this is not to provide a default.
        Call error() if the required option is not present.
        """
        option = self.get_option(opt)
        # Assumes the option's 'default' is set to None!
        if getattr(self.values, option.dest) is None:
            self.error("%s option not supplied" % option)

    def parse_args(self, args=None):
        """Process command-line options.

        Parameters:
            args - same as OptionParser.parse_args
        Return:
            True if all went well, False if not
        Post-conditions:
            If the return was True, logging levels and handlers
            are properly set for qualified name 'netlogger'.
            Otherwise, an error will be reported via the
            'error' method of the parser passed to the constructor.
        """
        if args is None:
            args = sys.argv[1:]
        self._add_options()
        options, pargs = optparse.OptionParser.parse_args(self, args)
        # Where and in what format to write logs
        if self._dmn:
            is_daemon = getattr(options, self.DEST_DAEMON)
        else:
            is_daemon = False
        logfile = getattr(options, self.DEST_LOG, None)
        logrot = getattr(options, self.DEST_ROT, None)
        if ((not logfile) or logfile == "-") and logrot:
            self.error("Log rotation requires a logfile")
        if logrot:
            if len(logrot) < 1:
                self.error("Bad log rotation interval, too short")
            tm_unit = logrot[-1].lower()
            if tm_unit not in ("h", "m", "d"):
                self.error("Bad log rotation unit '%s' " "not in m,h,d" % tm_unit)
            try:
                tm_interval = int(logrot[:-1])
            except ValueError:
                self.error("Log rotation value '%s' must be an integer" % logrot[:-1])
            do_logrot = True
            _tfrh = logging.handlers.TimedRotatingFileHandler
        else:
            do_logrot = False
        log = logging.getLogger(PROJECT_NAMESPACE)
        # Set handler and logger class
        handler = None
        if is_daemon:
            if logfile is None or logfile == "" or logfile == "-":  # missing/empty
                self.error("log file is required in daemon mode")
                return  # defensive
            else:
                # stderr and BP logs -> logfile
                setLoggerClass(BPLogger)
                logfile = logfile.strip()
                try:
                    if do_logrot:
                        handler = _tfrh(logfile, when=tm_unit, interval=tm_interval)
                    else:
                        handler = logging.FileHandler(logfile)
                except OSError as err:
                    self.error("Cannot open log file '{}': {}".format(logfile, err))
                sys.stderr = handler.stream
                handler.setFormatter(logging.Formatter("%(message)s"))
        else:
            if logfile is None or logfile == "":  # missing
                # Pretty-BP logs -> stderr
                setLoggerClass(PrettyBPLogger)
                handler = logging.StreamHandler()
            elif logfile.strip() == "-":  # empty
                # BP logs -> stderr
                setLoggerClass(BPLogger)
                handler = logging.StreamHandler()
            else:
                # BP logs -> logfile
                logfile = logfile.strip()
                setLoggerClass(BPLogger)
                try:
                    if do_logrot:
                        handler = _tfrh(logfile, when=tm_unit, interval=tm_interval)
                    else:
                        handler = logging.FileHandler(logfile)
                except OSError as err:
                    self.error("Cannot open log file '{}': {}".format(logfile, err))
                handler.setFormatter(logging.Formatter("%(message)s"))
        if handler:
            log.addHandler(handler)
        # Verbosity level
        quiet = getattr(options, self.DEST_QUIET, False)
        # delattr(options, self.DEST_QUIET)
        vb = getattr(options, self.DEST_VERBOSE, 0)
        # delattr(options, self.DEST_VERBOSE)
        if quiet and (vb > 0):
            self.error("quiet and verbosity options conflict")
            return  # defensive
        if quiet:
            log.setLevel(logging.CRITICAL + 1)
        else:
            log.setLevel(self.VBMAP[min(vb, len(self.VBMAP) - 1)])
        # Return remaining options and args to caller
        return options, pargs



__author__ = "Monte Goode"
__author__ = "Karan Vahi"

import time

from sqlalchemy import exc

from Pegasus.db.base_loader import BaseLoader
from Pegasus.db.schema import *


class DashboardLoader(BaseLoader):

    MAX_RETRIES = 10  # maximum number of retries in case of operational errors that arise because of database locked/connection dropped

    """Load into the Stampede Dashboard SQL schema through SQLAlchemy.

    Parameters:
      - connString {string,None*}: SQLAlchemy connection string.
        The general form of this is
          'dialect+driver://username:password@host:port/database'.
        See the SQLAlchemy docs for details.
        For sqlite, use 'sqlite:///foo.db' for a relative path and
        'sqlite:////path/to/foo.db' (four slashes) for an absolute one.
        When using MySQL, the general form will work, but the library
        expects the database to exist (ie: will not issue CREATE DB)
        but will populate an empty DB with tables/indexes/etc.
    """

    def __init__(
        self,
        connString,
        perf=False,
        batch=False,
        props=None,
        db_type=None,
        backup=False,
    ):
        """Init object

        @type   connString: string
        @param  connString: SQLAlchemy connection string - REQUIRED
        """
        super().__init__(
            connString,
            batch=batch,
            props=props,
            db_type=db_type,
            backup=backup,
            flush_every=1,
        )

        # "Case" dict to map events to handler methods
        self.eventMap = {
            "dashboard.wf.plan": self.workflow,
            #            'dashboard.wf.map.task_job' : self.task_map,
            "dashboard.xwf.start": self.workflowstate,
            "dashboard.xwf.end": self.workflowstate,
        }

        # Dicts for caching FK lookups
        self.wf_id_cache = {}
        self.root_wf_id_cache = {}

        # undocumented performance option
        self._perf = perf
        if self._perf:
            self._insert_time, self._insert_num = 0, 0
            self._start_time = time.time()

        # caches for batched events
        self._batch_cache = {
            "batch_events": [],
            "update_events": [],
            "host_map_events": [],
        }

    def process(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Get the BP dict from the controlling process and dispatch
        to the appropriate method per-event.
        """
        self.log.debug("process: %s", linedata)

        if not self._batch:
            self.check_connection()

        try:
            if self._perf:
                t = time.time()
                self.eventMap[linedata["event"]](linedata)
                self._insert_time += time.time() - t
                self._insert_num += 1
            else:
                self.eventMap[linedata["event"]](linedata)
        except KeyError:
            self.log.error('no handler for event type "%s" defined', linedata["event"])
        except exc.IntegrityError as e:
            # This is raised when an attempted insert violates the
            # schema (unique indexes, etc).
            self.log.error('Insert failed for event "%s" : %s', linedata["event"], e)
            self.session.rollback()
        except exc.OperationalError as e:
            self.log.error("Connection seemingly lost - attempting to refresh")
            self.session.rollback()
            self.check_connection()
            self.process(linedata)

        self.check_flush(increment=True)

    def linedataToObject(self, linedata, o):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.
        @type   o: instance of mapper class from stampede_dashboard_schema module.
        @param  o: Passed in by the appropriate event handler method.

        Takes the dict of BP linedata, assigns contents to the class o
        as attributes, and does any global type massaging like
        transforming dict strings to numeric types.
        """
        for k, v in linedata.items():
            if k == "level":
                continue

            # undot
            attr = k.replace(".", "_")

            attr_remap = {
                # workflow
                "xwf_id": "wf_uuid",
            }

            # remap attr names
            if attr in attr_remap:
                attr = attr_remap[attr]

            # sanitize argv input
            if attr == "argv":
                if v is not None:
                    v = v.replace("\\", "\\\\")
                    v = v.replace("'", "\\'")

            try:
                setattr(o, attr, v)
            except Exception:
                self.log.error("unable to process attribute %s with values: %s", k, v)

        # global type re-assignments
        if hasattr(o, "ts"):
            # make all timestamp values floats
            o.ts = float(o.ts)
        if hasattr(o, "restart_count") and o.restart_count is not None:
            o.restart_count = int(o.restart_count)
        return o

    #############################################
    # Methods to handle batching/flushing
    #############################################

    def hard_flush(self, batch_flush=True, retry=0):
        """
        @type   batch_flush: boolean
        @param  batch_flush: Defaults to true.  Is set to false
            when the batch commit hits and integrity error.

        Process queued inserts and flush/commit to the database.
        If the commit fails due to an integrity error, then method
        re-calls itself with setting batch_flush to False which
        causes each insert/object to be committed individually
        so all the "good" inserts can succeed.  This will increase
        the processing time of the batch with the bad data in it.
        """
        if not self._batch:
            return

        self.log.debug("Hard flush: batch_flush=%s", batch_flush)
        if retry == self.MAX_RETRIES + 1:
            # PM-1013 see if max retries is reached
            self.log.error(
                "Maximum number of retries reached for dashboard_loader.hard_flush() method %s"
                % self.MAX_RETRIES
            )
            raise RuntimeError(
                "Maximum number of retries reached for dashboard_loader.hard_flush() method %s"
                % self.MAX_RETRIES
            )

        retry = retry + 1
        self.check_connection()

        if self._perf:
            s = time.time()

        end_event = []

        for event in self._batch_cache["batch_events"]:
            if event.event == "dashboard.xwf.end":
                end_event.append(event)
            if batch_flush:
                self.session.add(event)
            else:
                self.individual_commit(event)

        for event in self._batch_cache["update_events"]:
            if batch_flush:
                self.session.merge(event)
            else:
                self.individual_commit(event, merge=True)

        try:
            self.session.commit()
        except exc.IntegrityError as e:
            self.log.error(
                "Integrity error on batch flush: %s - batch will need to be committed per-event which will take longer",
                e,
            )
            self.session.rollback()
        except exc.OperationalError as e:
            self.hard_flush(batch_flush=False, retry=retry)
            self.log.error(
                "Connection problem during commit: %s - reattempting batch", e
            )
            self.session.rollback()
            self.hard_flush(retry=retry)

        for host in self._batch_cache["host_map_events"]:
            self.map_host_to_job_instance(host)

        for ee in end_event:
            self.purgeCaches(ee)
        end_event = []

        # Clear all data structures here.
        for k in self._batch_cache.keys():
            self._batch_cache[k] = []

        self.session.commit()
        self.reset_flush_state()

        if self._perf:
            self.log.info("Hard flush duration: %s", (time.time() - s))

    #############################################
    # Methods to handle the various insert events
    #############################################
    def workflow(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a workflow insert event.
        """
        wf = self.linedataToObject(linedata, MasterWorkflow())
        self.log.debug("workflow: %s", wf)

        wf.timestamp = wf.ts

        is_root = True

        # for time being we don't track these. Karan
        #        if wf.root_xwf_id != wf.wf_uuid:
        #            is_root = False
        #            wf.root_wf_id = self.wf_uuid_to_id(wf.root_xwf_id)
        #
        #        if wf.parent_wf_id is not None:
        #            wf.parent_wf_id = self.wf_uuid_to_id(wf.parent_wf_id)

        # workflow inserts must be explicitly written to db whether
        # batching or not
        wf.commit_to_db(self.session)
        if is_root:
            wf.root_wf_id = self.wf_uuid_to_id(wf.root_xwf_id)
            wf.commit_to_db(self.session)
        if wf.root_wf_id is None:
            self.log.warn("Could not determine root_wf_id for event %s", wf)

    def workflowstate(self, linedata):
        """
        @type   linedata: dict
        @param  linedata: One line of BP data dict-ified.

        Handles a workflowstate insert event.
        """
        wfs = self.linedataToObject(linedata, MasterWorkflowstate())
        self.log.debug("workflowstate: %s", wfs)

        state = {
            "dashboard.xwf.start": "WORKFLOW_STARTED",
            "dashboard.xwf.end": "WORKFLOW_TERMINATED",
        }

        wfs.wf_id = self.wf_uuid_to_id(wfs.wf_uuid)
        wfs.timestamp = wfs.ts
        wfs.state = state[wfs.event]

        if self._batch:
            self._batch_cache["batch_events"].append(wfs)
        else:
            wfs.commit_to_db(self.session)
            if wfs.event == "dashboard.xwf.end":
                self.purgeCaches(wfs)

    ####################################
    # DB helper/lookup/caching functions
    ####################################
    def wf_uuid_to_id(self, wf_uuid):
        """
        @type   wf_uuid: string
        @param  wf_uuid: wf_uuid string from BP logs

        Attempts to retrieve a workflow wf_id PK/FK from cache.  If
        not in cache, retrieve from st_workflow table in DB and cache.
        Cuts down on DB queries during insert processing.
        """
        if wf_uuid not in self.wf_id_cache:
            query = self.session.query(MasterWorkflow).filter(
                MasterWorkflow.wf_uuid == wf_uuid
            )
            try:
                self.wf_id_cache[wf_uuid] = query.one().wf_id
            except orm.exc.MultipleResultsFound as e:
                self.log.error("Multiple wf_id results for wf_uuid %s : %s", wf_uuid, e)
                return None
            except orm.exc.NoResultFound as e:
                self.log.error("No wf_id results for wf_uuid %s : %s", wf_uuid, e)
                return None

        return self.wf_id_cache[wf_uuid]

    def wf_uuid_to_root_id(self, wf_uuid):
        """
        @type   wf_uuid: string
        @param  wf_uuid: wf_uuid string from BP logs

        Attempts to retrieve a root workflow wf_id PK/FK from cache.  If
        not in cache, retrieve from st_workflow table in DB and cache.
        Cuts down on DB queries during insert processing.
        """
        if wf_uuid not in self.root_wf_id_cache:
            query = self.session.query(Workflow).filter(
                MasterWorkflow.wf_uuid == wf_uuid
            )
            try:
                self.root_wf_id_cache[wf_uuid] = query.one().root_wf_id
            except orm.exc.MultipleResultsFound as e:
                self.log.error("Multiple wf_id results for wf_uuid %s : %s", wf_uuid, e)
                return None
            except orm.exc.NoResultFound as e:
                self.log.error("No wf_id results for wf_uuid %s : %s", wf_uuid, e)
                return None

        return self.root_wf_id_cache[wf_uuid]

    def purgeCaches(self, wfs):
        """
        @type   wfs: class instance of stampede_schema.Workflowstate
        @param  wfs: Workflow state object from an end event.

        Purges information from the lookup caches after a workflow.end
        event has been recieved.
        """
        self.log.debug("Purging caches for: %s", wfs.wf_uuid)

        """
        for k,v in self.wf_id_cache.items():
            if k == wfs.wf_uuid:
                del self.wf_id_cache[k]

        for k,v in self.root_wf_id_cache.items():
            if k == wfs.wf_uuid:
                del self.root_wf_id_cache[k]
        """
        self.purgeCache(self.wf_id_cache, wfs.wf_uuid)
        self.purgeCache(self.root_wf_id_cache, wfs.wf_uuid)

    def purgeCache(self, cache, key):
        """
        Removes from a cache an entry matching a key id
        :param cache:
        :param key_id:
        :return:
        """

        if key in cache:
            del cache[key]

    ################
    # Cleanup, etc
    ################

    def finish(self):
        if self._batch:
            self.log.info("Executing final flush")
            self.hard_flush()
        self.disconnect()
        if self._perf:
            run_time = time.time() - self._start_time
            self.log.info(
                "Loader performance: insert_time=%s, insert_num=%s, "
                "total_time=%s, run_time_delta=%s, mean_time=%s",
                self._insert_time,
                self._insert_num,
                run_time,
                run_time - self._insert_time,
                self._insert_time / self._insert_num,
            )



import os
import pwd


class NoSuchUser(Exception):
    pass


class User:
    def __init__(self, uid, gid, username, homedir):
        self.uid = uid
        self.gid = gid
        self.username = username
        self.homedir = homedir

    def get_pegasus_dir(self):
        return os.path.join(self.homedir, ".pegasus")

    def get_ensembles_dir(self):
        return os.path.join(self.homedir, ".pegasus", "ensembles")

    def get_master_db(self):
        return os.path.join(self.homedir, ".pegasus", "workflow.db")

    def get_master_db_url(self):
        return "sqlite:///%s" % self.get_master_db()


def __user_from_pwd(pw):
    return User(pw.pw_uid, pw.pw_gid, pw.pw_name, pw.pw_dir)


def get_user_by_uid(uid):
    try:
        pw = pwd.getpwuid(uid)
        return __user_from_pwd(pw)
    except KeyError:
        raise NoSuchUser(uid)


def get_user_by_username(username):
    try:
        pw = pwd.getpwnam(username)
        return __user_from_pwd(pw)
    except KeyError:
        raise NoSuchUser(username)
    except TypeError:
        raise NoSuchUser(username)



import os
import zipfile

from Pegasus.tools import properties

PROPERTIES_NAME = "pegasus.properties"


class BundleException(Exception):
    pass


class Bundle:
    def __init__(self, filename):
        self.filename = filename

        if not os.path.isfile(filename):
            raise BundleException("No such file")

        # Verify that the bundle is a valid zip file
        if not zipfile.is_zipfile(filename):
            raise BundleException("Bundle is not a valid zip file")

        # open the bundle
        self.zipfile = zipfile.ZipFile(filename, "r")

        # Verify that the bundle does not contain any harmful paths
        for name in self.zipfile.namelist():
            if name.startswith("/") or ".." in name:
                raise BundleException("Invalid bundle entry: %s" % name)

        # Verify that the bundle contains a manifest file
        if not self.contains(PROPERTIES_NAME):
            raise BundleException("Bundle does not contain %s" % PROPERTIES_NAME)

        # Open the properties entry
        self.properties = properties.parse_properties(
            self.zipfile.open(PROPERTIES_NAME, "r")
        )

    def contains(self, name):
        # getinfo raises a KeyError if the entry does not exist
        try:
            self.zipfile.getinfo(name)
            return True
        except KeyError:
            return False

    def _ensure_valid_file(self, propname):
        filename = self.properties.get(propname)
        if filename is None:
            return

        if not self.contains(filename) and not os.path.isfile(filename):
            raise BundleException(
                "Invalid {}: No such file: {}".format(propname, filename)
            )

    def verify(self):
        # Make sure the dax property is set
        daxfile = self.properties.get("pegasus.dax.file")
        if daxfile is None:
            raise BundleException("Bundle properties does not contain pegasus.dax.file")

        # Make sure these file properties, if they are set, point to valid
        # files in either the zip file or the file system
        self._ensure_valid_file("pegasus.dax.file")
        self._ensure_valid_file("pegasus.catalog.replica.file")
        self._ensure_valid_file("pegasus.catalog.transformation.file")
        self._ensure_valid_file("pegasus.catalog.site.file")

    def get_properties(self):
        return self.properties

    def unpack(self, dirname):
        self.zipfile.extractall(dirname)



"""
:mod:`kickstart` exposes an API to serialize and deserialize Pegasus's kickstart file.

Basic Usage::

    >>> from Pegasus import kickstart
    >>> kickstart.loads("... ")
    ...

    >>> print(kickstart.dumps( ... ))
    ' ... '

.. moduleauthor:: Mats Rynge <rynge@isi.edu>
.. moduleauthor:: Rajiv Mayani <mayani@isi.edu>
"""

from typing import Dict, TextIO

__all__ = (
    "load",
    "loads",
    "dump",
    "dumps",
)


def load(fp: TextIO, *args, **kwargs) -> Dict:
    """
    Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a Kickstart document) to a Python object.

    [extended_summary]

    :param fp: [description]
    :type fp: TextIO
    :return: [description]
    :rtype: Dict
    """


def loads(s: str, *args, **kwargs) -> Dict:
    """
    Deserialize ``s`` (a ``str``, ``bytes`` or ``bytearray`` instance containing a Kickstart document) to a Python object.

    [extended_summary]

    :param s: [description]
    :type s: str
    :return: [description]
    :rtype: Dict
    """


def dump(obj: Dict, fp: TextIO, *args, **kwargs) -> None:
    """
    Serialize ``obj`` as a Kickstart formatted stream to ``fp`` (a ``.write()``-supporting file-like object).

    [extended_summary]

    :param obj: [description]
    :type obj: Dict
    :param fp: [description]
    :type fp: TextIO
    :return: [description]
    :rtype: NoReturn
    """


def dumps(obj: Dict, *args, **kwargs) -> str:
    """
    Serialize ``obj`` to a Kickstart formatted ``str``.

    [extended_summary]

    :param obj: [description]
    :type obj: Dict
    :return: [description]
    :rtype: str
    """



#!/usr/bin/env python3

"""
Pegasus Metadata is a tool to query metadata collected by Pegasus workflows.

Usage: pegasus-metadata [-h] [-v] [-c] {task,file,workflow} ... submit_dir
"""

##
#  Copyright 2007-2012 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##


import argparse
import logging
import sys
from pathlib import Path

from Pegasus.db import connection
from Pegasus.db.connection import ConnectionError, DBType
from Pegasus.service.monitoring.queries import (
    StampedeDBNotFoundError,
    StampedeWorkflowQueries,
)
from Pegasus.tools import utils


def configure_logging(verbosity=0):
    verbosity = min(3, verbosity)

    log_levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]

    utils.configureLogging(level=log_levels[verbosity])


def get_workflow_uuid(submit_dir):
    bdump_yml = Path(submit_dir) / "braindump.yml"
    bdump_txt = Path(submit_dir) / "braindump.txt"

    if bdump_yml.exists() is False and bdump_txt.exists() is False:
        raise ValueError("Not a valid workflow submit directory: %r" % submit_dir)

    braindump = utils.slurp_braindb(submit_dir)
    return braindump["root_wf_uuid"], braindump["wf_uuid"]


def get_workflow_uri(submit_dir):
    return connection.url_by_submitdir(submit_dir, DBType.WORKFLOW)


def render_metas(metas, indent=""):
    if not metas:
        print("%sNo metadata found" % indent)
        return

    max_key_len = 0

    for meta in metas:
        max_key_len = max(max_key_len, len(meta.key))

    max_key_len += 1

    for meta in metas:
        print("{}{}: {}".format(indent, meta.key.ljust(max_key_len), meta.value))


def workflow_metadata(recursive=False, submit_dir=".", *args, **kwargs):
    logging.debug("workflow_metadata")

    try:
        root_wf_uuid, wf_uuid = get_workflow_uuid(submit_dir)
        logging.debug("Workflow UUID: %s" % wf_uuid)
        db_uri = get_workflow_uri(submit_dir)

        queries = StampedeWorkflowQueries(db_uri)

        if recursive:
            wfs = queries.get_workflows(
                root_wf_uuid, order="w.root_wf_id, w.parent_wf_id"
            )

            for wf in wfs.records:
                print("Workflow %s" % wf.wf_uuid)
                workflow_metas = queries.get_workflow_meta(wf_uuid).records
                render_metas(workflow_metas, "    ")

        else:
            workflow_metas = queries.get_workflow_meta(wf_uuid).records
            print("Workflow %s" % wf_uuid)
            render_metas(workflow_metas, "    ")

    except ValueError as e:
        logging.error(e)
        sys.exit(1)

    except ConnectionError as e:
        logging.error(e)
        sys.exit(2)

    except StampedeDBNotFoundError as e:
        logging.error(e)
        sys.exit(3)


def task_metadata(abs_task_id=None, submit_dir=".", *args, **kwargs):
    logging.debug("task_metadata")

    if not abs_task_id:
        logging.error("task_id is required")
        sys.exit(1)

    try:
        root_wf_uuid, wf_uuid = get_workflow_uuid(submit_dir)
        logging.debug("Workflow UUID: %s" % wf_uuid)
        db_uri = get_workflow_uri(submit_dir)

        queries = StampedeWorkflowQueries(db_uri)

        logging.debug("Get task metadata for abs_task_id %s" % abs_task_id)
        workflow = queries.get_workflow_tasks(
            wf_uuid, query="t.abs_task_id == %r" % abs_task_id
        )

        if workflow.total_filtered == 0:
            raise ValueError("Invalid task_name %r" % abs_task_id)

        task_id = workflow.records[0].task_id
        task_metas = queries.get_task_meta(task_id).records

        print("Task", abs_task_id)
        render_metas(task_metas, "    ")

    except ValueError as e:
        logging.error(e)
        sys.exit(1)

    except ConnectionError as e:
        logging.error(e)
        sys.exit(2)

    except StampedeDBNotFoundError as e:
        logging.error(e)
        sys.exit(3)


def file_metadata(
    file_name=None, list_files=False, trace=False, submit_dir=".", *args, **kwargs
):
    logging.debug("task_metadata")

    if trace and not file_name:
        logging.error("file_name is required when trace is True")
        sys.exit(1)

    if not file_name:
        list_files = True
        logging.info("file_name not provided, will list metadata for all files")

    try:
        root_wf_uuid, wf_uuid = get_workflow_uuid(submit_dir)
        logging.debug("Workflow UUID: %s" % wf_uuid)
        db_uri = get_workflow_uri(submit_dir)

        queries = StampedeWorkflowQueries(db_uri)

        if list_files:
            logging.debug("Get file metadata for all files")
            workflow_files = queries.get_workflow_files(wf_uuid)

            if workflow_files.total_filtered == 0:
                print("No files found")

        else:
            logging.debug("Get file metadata for lfn %r" % file_name)

            workflow_files = queries.get_workflow_files(
                wf_uuid, query="l.lfn == %r" % file_name
            )

            if workflow_files.total_filtered == 0:
                raise ValueError("Invalid file %r" % file_name)

        if trace:
            wf_file = workflow_files.records.values()[0]
            wf_id = wf_file.extras.wf_id
            task_id = wf_file.extras.task_id

            # Get workflow information
            root_wf = None
            wf = queries.get_workflow(wf_id)

            # If workflow is hierarchical workflow, get root workflow information
            if wf_id != wf.root_wf_id:
                root_wf = queries.get_workflow(wf.root_wf_id)

            if root_wf:
                root_wf_metas = queries.get_workflow_meta(root_wf.wf_id).records
                print("Root Workflow %s" % root_wf.wf_uuid)
                render_metas(root_wf_metas, "    ")
                print()

            wf_metas = queries.get_workflow_meta(wf_id).records
            print("Workflow %s" % wf.wf_uuid)
            render_metas(wf_metas, "    ")
            print()

            task = queries.get_task(task_id)
            task_metas = queries.get_task_meta(task_id).records
            print("Task %s" % task.abs_task_id)
            render_metas(task_metas, "    ")
            print()

        for wf_file in workflow_files.records:
            print("File %s" % wf_file.lfn)
            render_metas(wf_file.meta, "    ")

    except ValueError as e:
        logging.error(e)
        sys.exit(1)

    except ConnectionError as e:
        logging.error(e)
        sys.exit(2)

    except StampedeDBNotFoundError as e:
        logging.error(e)
        sys.exit(3)


def main():
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "-v", "--verbose", default=0, action="count", help="Logging verbosity"
    )
    parent_parser.add_argument(
        "submit_dir", nargs="?", default=".", help="Workflow submit directory"
    )

    parser = argparse.ArgumentParser(description="Pegasus Metadata Query Tool")
    sub_parser = parser.add_subparsers(
        title="Metadata types", description="Types of metadata that can be queried"
    )

    # Workflow Metadata Options
    workflow = sub_parser.add_parser("workflow", parents=[parent_parser])
    workflow.add_argument("-r", "--recursive", default=False, action="store_true")
    workflow.set_defaults(func=workflow_metadata)

    # Task Metadata Options
    task = sub_parser.add_parser("task", parents=[parent_parser])
    task.add_argument("-i", "--task-id", dest="abs_task_id", required=True)
    task.set_defaults(func=task_metadata)

    # File Metadata Options
    file = sub_parser.add_parser("file", parents=[parent_parser])
    file.add_argument("-n", "--file-name")
    file.add_argument(
        "-l", "--list", default=False, action="store_true", dest="list_files"
    )
    file.add_argument("-t", "--trace", default=False, action="store_true")
    file.set_defaults(func=file_metadata)

    args = parser.parse_args(sys.argv[1:])

    configure_logging(args.verbose)
    args.func(**vars(args))


if __name__ == "__main__":
    main()



from collections import defaultdict
from io import StringIO
from typing import TextIO

from Pegasus import yaml
from Pegasus.api.errors import PegasusError
from Pegasus.api.replica_catalog import File
from Pegasus.api.workflow import (
    Job,
    SubWorkflow,
    Workflow,
    _JobDependency,
    _LinkType,
    _Use,
)
from Pegasus.replica_catalog import _to_rc
from Pegasus.site_catalog import _to_sc
from Pegasus.transformation_catalog import _to_tc

"""
:mod:`workflow` exposes an API to serialize and deserialize Pegasus's workflow file.

Basic Usage::

    >>> from Pegasus import workflow
    >>> workflow.loads("... ")
    ...

    >>> print(workflow.dumps( ... ))
    ' ... '

.. moduleauthor:: Ryan Tanaka <tanaka@isi.edu>
.. moduleauthor:: Rajiv Mayani <mayani@isi.edu>
"""


__all__ = (
    "load",
    "loads",
    "dump",
    "dumps",
)


def _to_wf(d: dict) -> Workflow:
    """Convert dict to Workflow

    :param d: Workflow represented as a dict
    :type d: dict
    :raises PegasusError: encountered error parsing
    :return: a Workflow object based on d
    :rtype: Workflow
    """

    try:
        #
        wf = Workflow(d["name"], infer_dependencies=False)

        # add rc
        if "replicaCatalog" in d:
            wf.replica_catalog = _to_rc(d["replicaCatalog"])

        # add tc
        if "transformationCatalog" in d:
            wf.transformation_catalog = _to_tc(d["transformationCatalog"])

        # add sc
        if "siteCatalog" in d:
            wf.site_catalog = _to_sc(d["siteCatalog"])

        # add jobs
        for j in d["jobs"]:
            # create appropriate job based on type
            if j["type"] == "job":
                job = Job(
                    j["name"],
                    _id=j["id"],
                    node_label=j.get("nodeLabel"),
                    namespace=j.get("namespace"),
                    version=j.get("version"),
                )
            elif j["type"] in {"pegasusWorkflow", "condorWorkflow"}:
                f = File(j["file"])

                is_planned = False if j["type"] == "pegasusWorkflow" else True

                job = SubWorkflow(
                    f, is_planned, _id=j["id"], node_label=j.get("nodeLabel")
                )

            else:
                raise ValueError

            # add args
            args = list()
            for a in j["arguments"]:
                args.append(a)

            job.args = args

            # add uses
            uses = set()
            for u in j["uses"]:
                f = File(u["lfn"], size=u.get("size"))
                try:
                    f.metadata = u["metadata"]
                except KeyError:
                    pass

                uses.add(
                    _Use(
                        f,
                        getattr(_LinkType, u["type"].upper()),
                        stage_out=u.get("stageOut"),
                        register_replica=u.get("registerReplica"),
                        bypass_staging=u.get("bypass"),
                    )
                )

            job.uses = uses

            # set stdin
            if "stdin" in j:
                for u in job.uses:
                    if u.file.lfn == j["stdin"]:
                        job.stdin = u.file
                        break

            # set stdout
            if "stdout" in j:
                for u in job.uses:
                    if u.file.lfn == j["stdout"]:
                        job.stdout = u.file
                        break

            # set stderr
            if "stderr" in j:
                for u in job.uses:
                    if u.file.lfn == j["stderr"]:
                        job.stderr = u.file
                        break

            # add profiles
            if j.get("profiles"):
                job.profiles = defaultdict(dict, j.get("profiles"))

            # add metadata
            if j.get("metadata"):
                job.metadata = j.get("metadata")

            # add hooks
            if j.get("hooks"):
                job.hooks = defaultdict(list, j.get("hooks"))

            # add job to wf
            wf.add_jobs(job)

        # add dependencies
        if d.get("jobDependencies"):
            dependencies = defaultdict(_JobDependency)
            for item in d.get("jobDependencies"):
                dependencies[item["id"]] = _JobDependency(
                    item["id"], {child for child in item["children"]}
                )

            wf.dependencies = dependencies

        # add profiles
        if d.get("profiles"):
            wf.profiles = defaultdict(dict, d.get("profiles"))

        # add metadata
        if d.get("metadata"):
            wf.metadata = d.get("metadata")

        # add hooks
        if d.get("hooks"):
            wf.hooks = defaultdict(list, d.get("hooks"))

        return wf
    except (KeyError, ValueError):
        raise PegasusError("error parsing {}".format(d))


def load(fp: TextIO, *args, **kwargs) -> Workflow:
    """
    Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a Workflow document) to a :py:class:`~Pegasus.api.workflow.Workflow` object.

    :param fp: file like object to load from
    :type fp: TextIO
    :return: deserialized Workflow object
    :rtype: Workflow
    """
    return _to_wf(yaml.load(fp))


def loads(s: str, *args, **kwargs) -> Workflow:
    """
    Deserialize ``s`` (a ``str``, ``bytes`` or ``bytearray`` instance containing a Workflow document) to a :py:class:`~Pegasus.api.workflow.Workflow` object.

    :param s: string to load from
    :type s: str
    :return: deserialized Workflow object
    :rtype: Workflow
    """
    return _to_wf(yaml.load(s))


def dump(obj: Workflow, fp: TextIO, _format="yml", *args, **kwargs) -> None:
    """
    Serialize ``obj`` as a :py:class:`~Pegasus.api.worklfow.Workflow` formatted stream to ``fp`` (a ``.write()``-supporting file-like object).

    :param obj: Workflow to serialize
    :type obj: Workflow
    :param fp: file like object to serialize to
    :type fp: TextIO
    :param _format: format to write to if fp does not have an extension; can be one of ["yml" | "yaml" | "json"], defaults to "yml"
    :type _format: str
    :rtype: NoReturn
    """
    obj.write(fp, _format=_format)


def dumps(obj: Workflow, _format="yml", *args, **kwargs) -> str:
    """
    Serialize ``obj`` to a :py:class:`~Pegasus.api.workflow.Workflow` formatted ``str``.

    :param obj: Workflow to serialize
    :type obj: Workflow
    :param _format: format to write to if fp does not have an extension; can be one of ["yml" | "yaml" | "json"], defaults to "yml"
    :type _format: str
    :return: Workflow serialized as a string
    :rtype: str
    """
    with StringIO() as s:
        obj.write(s, _format=_format)
        s.seek(0)
        return s.read()
